import numpy as np

from mosamaticdesktop.tasks.musclefatsegmentationtask.model import Model


class TensorFlowModel(Model):
    def __init__(self) -> None:
        super(TensorFlowModel, self).__init__()
        self._model = None

    def load(self, modelFilePath: str) -> None:
        import tensorflow as tf
        self._model = tf.keras.models.load_model(modelFilePath, compile=False)

    def predict(self, input: np.array) -> np.array:
        if self._model:
            return self._model.predict(input)
        raise RuntimeError('Model not loaded yet')


from mosamaticdesktop.widgets.dockwidget import DockWidget
from mosamaticdesktop.widgets.filesettreeview import FileSetTreeView


class DataDockWidget(DockWidget):
    def __init__(self, title: str) -> None:
        super(DataDockWidget, self).__init__(title)
        self._treeView = None
        self.initUi()

    def initUi(self) -> None:
        self._treeView = FileSetTreeView()
        self.setWidget(self._treeView)
        self.setMinimumHeight(200)

    def treeView(self) -> FileSetTreeView:
        return self._treeView


import os
import sys
import subprocess

from PySide6.QtCore import QPoint
from PySide6.QtWidgets import QMenu, QTreeView

from mosamaticdesktop.widgets.fileitem import FileItem

PLATFORM = sys.platform


class FileItemMenu(QMenu):
    def __init__(self, treeView: QTreeView, fileItem: FileItem, position: QPoint) -> None:
        super(FileItemMenu, self).__init__()
        self._treeView = treeView
        self._fileItem = fileItem
        self._position = position
        action = self.addAction('Find File in Finder/Explorer')
        action.triggered.connect(self.findFileInFinderOrExplorer)

    def findFileInFinderOrExplorer(self):
        if PLATFORM == 'darwin':
            subprocess.run(['open', '-R', self._fileItem.file().path()])
        elif PLATFORM in ['win32', 'cygwin']:
            subprocess.run(['explorer', '/select,', os.path.normpath(self._fileItem.file().path())])
        else:
            pass

    def show(self):
        self.exec_(self._position)


import os
import numpy as np

from mosamaticdesktop.tasks.task import Task
from mosamaticdesktop.utils import convertDicomToNumPyArray
from mosamaticdesktop.utils import convertNumPyArrayToPngImage
from mosamaticdesktop.utils import isDicomFile
from mosamaticdesktop.logger import Logger

LOGGER = Logger()


class CreatePngFromDicomFileTask(Task):
    def __init__(self) -> None:
        super(CreatePngFromDicomFileTask, self).__init__()
        self.addDescriptionParameter(
            name='description',
            description=f'Create PNGs From DICOM Files',
        )
        self.addFileSetParameter(
            name='inputFileSetName',
            labelText='Input File Set',
        )
        self.addIntegerParameter(
            name='windowLevel',
            labelText='Window Level',
            minimum=0,
            maximum=2000,
            step=1,
            defaultValue=50,
        )
        self.addIntegerParameter(
            name='windowWidth',
            labelText='Window Width',
            minimum=0,
            maximum=2000,
            step=1,
            defaultValue=400,
        )
        self.addPathParameter(
            name='outputFileSetPath',
            labelText='Output File Set Path',
        )
        self.addTextParameter(
            name='outputFileSetName',
            labelText='Output File Set Name',
            optional=True,
        )
        self.addBooleanParameter(
            name='overwriteOutputFileSet',
            labelText='Overwrite Output File Set',
            defaultValue=True,
        )

    def execute(self) -> None:

        # Get parameters needed for this task
        inputFileSetName = self.parameter(name='inputFileSetName').value()
        inputFileSet = self.dataManager().fileSetByName(name=inputFileSetName)
        windowLevel = self.parameter(name='windowLevel').value()
        windowWidth = self.parameter(name='windowWidth').value()
        outputFileSetPath = self.parameter('outputFileSetPath').value()
        outputFileSetName = self.parameter('outputFileSetName').value()
        if outputFileSetName is None:
            outputFileSetName = self.generateTimestampForFileSetName(name=inputFileSetName)
        overwriteOutputFileSet = self.parameter('overwriteOutputFileSet').value()
        outputFileSetPath = os.path.join(outputFileSetPath, outputFileSetName)
        if overwriteOutputFileSet:
            if os.path.isdir(outputFileSetPath):
                shutil.rmtree(outputFileSetPath)
        os.makedirs(outputFileSetPath, exist_ok=True)

        step = 0
        files = inputFileSet.files()
        nrSteps = len(files)
        for file in files:
            if isDicomFile(file.path()):
                numpyArray = convertDicomToNumPyArray(file.path(), windowLevel=windowLevel, windowWidth=windowWidth)
                pngImageFileName = os.path.split(file.path())[1] + '.png'
                convertNumPyArrayToPngImage(
                    numpyArrayFilePathOrObject=numpyArray, 
                    colorMap=None,
                    outputDirectoryPath=outputFileSetPath,
                    pngImageFileName=pngImageFileName,
                )
                LOGGER.info(f'{file.path()} created {pngImageFileName}')
            self.updateProgress(step=step, nrSteps=nrSteps)
            step += 1
        self.dataManager().createFileSet(fileSetPath=outputFileSetPath)
        LOGGER.info('Finished')


import os

from typing import List

from PySide6.QtCore import QObject, Signal

from mosamaticdesktop.data.session import Session
from mosamaticdesktop.data.fileset import FileSet
from mosamaticdesktop.data.models.filesetmodel import FileSetModel
from mosamaticdesktop.data.file import File
from mosamaticdesktop.data.models.filemodel import FileModel
from mosamaticdesktop.data.filecontentcache import FileContentCache
from mosamaticdesktop.singleton import singleton
from mosamaticdesktop.logger import Logger

LOGGER = Logger()


@singleton
class DataManager:
    class DataManagerUpdatedSignal(QObject):
        updated = Signal(bool)

    def __init__(self) -> None:
        self._signal = self.DataManagerUpdatedSignal()

    def signal(self):
        return self._signal

    def createFile(self, filePath: str) -> FileSet:
        with Session() as session:
            fileSetPath = os.path.split(filePath)[0]
            fileSetName = fileSetPath.split(os.path.sep)[-1]
            fileSetModel = FileSetModel(name=fileSetName, path=fileSetPath)
            session.add(fileSetModel)
            fileName = os.path.split(filePath)[1]
            fileModel = FileModel(name=fileName, path=filePath, fileSetModel=fileSetModel)
            session.add(fileModel)
            session.commit()
            fileSet = FileSet(model=fileSetModel)
        self._signal.updated.emit(True)
        return fileSet


    def createFileSet(self, fileSetPath: str) -> FileSet:
        fileSetName = os.path.split(fileSetPath)[-1]
        with Session() as session:
            fileSetModel = FileSetModel(name=fileSetName, path=fileSetPath)
            session.add(fileSetModel)
            session.commit()
            fileSet = FileSet(model=fileSetModel)
            nrFiles = 0
            for fileName in os.listdir(fileSetPath):
                if fileName.startswith('.'):
                    continue
                filePath = os.path.join(fileSetPath, fileName)
                fileModel = FileModel(name=fileName, path=filePath, fileSetModel=fileSetModel)
                session.add(fileModel)
                session.commit()
                file = File(model=fileModel)
                fileSet.addFile(file)
                nrFiles += 1
        self._signal.updated.emit(True)
        return fileSet
    
    def fileSet(self, id: str) -> FileSet:
        with Session() as session:
            fileSetModel = session.get(FileSetModel, id)
            if fileSetModel:
                fileSet = FileSet(model=fileSetModel)
                return fileSet
        return None
    
    def fileSetByName(self, name: str) -> FileSet:
        with Session() as session:
            fileSetModel = session.query(FileSetModel).filter_by(name=name).one()
            if fileSetModel:
                fileSet = FileSet(model=fileSetModel)
                return fileSet
        return None
    
    def fileSets(self) -> List[FileSet]:
        with Session() as session:
            fileSetModels = session.query(FileSetModel).all()
            fileSets = []
            for fileSetModel in fileSetModels:
                fileSet = FileSet(model=fileSetModel)
                fileSets.append(fileSet)
        return fileSets

    def updateFileSet(self, fileSet: FileSet) -> FileSet:
        with Session() as session:
            fileSetModel = session.get(FileSetModel, fileSet.id())
            if fileSetModel:
                fileSetModel.name = fileSet.name()
                session.commit()
                return FileSet(model=fileSetModel)
        self._signal.updated.emit(True)
        return None
    
    def deleteFileSet(self, fileSet: FileSet) -> None:
        with Session() as session:
            fileSetModel = session.get(FileSetModel, fileSet.id())
            session.delete(fileSetModel)
            session.commit()
            cache = FileContentCache()
            for file in fileSet.files():
                cache.remove(file.id())
        self._signal.updated.emit(True)

    def deleteAllFileSets(self) -> None:
        with Session() as session:
            cache = FileContentCache()
            fileSetModels = session.query(FileSetModel).all()
            for fileSetModel in fileSetModels:
                for fileModel in fileSetModel.fileModels:
                    cache.remove(fileModel.id)
                session.delete(fileSetModel)
            session.commit()
        self._signal.updated.emit(True)



import os
import json
import shutil
import zipfile
import pydicom
import pydicom.errors
import numpy as np

from typing import List, Any

from mosamaticdesktop.tasks.task import Task
from mosamaticdesktop.tasks.musclefatsegmentationtask.tensorflowmodel import TensorFlowModel
from mosamaticdesktop.data.file import File
from mosamaticdesktop.logger import Logger
from mosamaticdesktop.utils import getPixelsFromDicomObject, convertLabelsTo157
from mosamaticdesktop.utils import normalizeBetween, Configuration
from mosamaticdesktop.utils import readFromCache, writeToCache

LOGGER = Logger()


class MuscleFatSegmentationTask(Task):
    ARGMAX = 0
    PROBABILITIES = 1

    def __init__(self) -> None:
        super(MuscleFatSegmentationTask, self).__init__()        
        self.addDescriptionParameter(
            name='description',
            description='Extract muscle and fat regions from L3 images'
        )
        self.addFileSetParameter(
            name='inputFileSetName',
            labelText='Input File Set',
        )
        self.addFileSetParameter(
            name='tensorFlowModelFileSetName',
            labelText='TensorFlow Model File Set',
        )
        self.addOptionGroupParameter(
            name='mode',
            labelText='Mode',
            defaultValue='ARGMAX',
            options=['ARGMAX', 'PROBABILITIES'],
        )
        self.addPathParameter(
            name='outputFileSetPath',
            labelText='Output File Set Path',
        )
        self.addTextParameter(
            name='outputFileSetName',
            labelText='Output File Set Name',
            optional=True,
        )
        self.addBooleanParameter(
            name='overwriteOutputFileSet',
            labelText='Overwrite Output File Set',
            defaultValue=True,
        )

    def loadModelFiles(self, files: List[File]) -> List[Any]:
        tfLoaded = False
        configuration = Configuration()
        for file in files:
            if file.name() == 'model.zip':
                content = readFromCache(file=file)
                if not content:
                    if not tfLoaded:
                        import tensorflow as tf # Only load TensorFlow package if necessary (takes some time)
                        tfLoaded = True
                    tensorFlowModelFileDirectory = configuration.taskConfigSubDirectory(taskName=__class__.__name__, dirName='tensorFlowModelFiles')
                    with zipfile.ZipFile(file.path()) as zipObj:
                        zipObj.extractall(path=tensorFlowModelFileDirectory)
                    tensorFlowModel = TensorFlowModel()
                    tensorFlowModel.load(modelFilePath=tensorFlowModelFileDirectory)
                    content = writeToCache(file=file, fileObject=tensorFlowModel)
                tensorFlowModel = content.fileObject()
            elif file.name() == 'contour_model.zip':
                content = readFromCache(file=file)
                if not content:
                    if not tfLoaded:
                        import tensorflow as tf # Only load TensorFlow package if necessary (takes some time)
                        tfLoaded = True
                    tensorFlowModelFileDirectory = configuration.taskConfigSubDirectory(taskName=__class__.__name__, dirName='tensorFlowModelFiles')
                    with zipfile.ZipFile(file.path()) as zipObj:
                        zipObj.extractall(path=tensorFlowModelFileDirectory)
                    tensorFlowContourModel = TensorFlowModel()
                    tensorFlowContourModel.load(modelFilePath=tensorFlowModelFileDirectory)
                    content = writeToCache(file=file, fileObject=tensorFlowContourModel)
                tensorFlowContourModel = content.fileObject()
            elif file.name() == 'params.json':
                content = readFromCache(file=file)
                if not content:
                    with open(file.path(), 'r') as f:
                        parameters = json.load(f)
                        content = writeToCache(file=file, fileObject=parameters)
                parameters = content.fileObject()
            else:
                pass
        return [tensorFlowModel, tensorFlowContourModel, parameters]

    def predictContour(self, contourModel, sourceImage, parameters):
        ct = np.copy(sourceImage)
        ct = normalizeBetween(ct, parameters['min_bound_contour'], parameters['max_bound_contour'])
        img2 = np.expand_dims(ct, 0)
        img2 = np.expand_dims(img2, -1)
        pred = contourModel.predict([img2])
        predSqueeze = np.squeeze(pred)
        pred_max = predSqueeze.argmax(axis=-1)
        mask = np.uint8(pred_max)
        return mask
    
    def execute(self) -> None:

        # Get input fileset
        inputFileSetName = self.parameter('inputFileSetName').value()
        inputFileSet = self.dataManager().fileSetByName(name=inputFileSetName)
        tensorFlowModelFileSetName = self.parameter('tensorFlowModelFileSetName').value()
        tensorFlowModelFileSet = self.dataManager().fileSetByName(tensorFlowModelFileSetName)
        # Setup output fileset path
        outputFileSetName = self.parameter('outputFileSetName').value()
        if outputFileSetName is None:
            outputFileSetName = self.generateTimestampForFileSetName(name=inputFileSetName)
        outputFileSetPath = self.parameter('outputFileSetPath').value()
        outputFileSetPath = os.path.join(outputFileSetPath, outputFileSetName)

        overwriteOutputFileSet = self.parameter('overwriteOutputFileSet').value()
        LOGGER.info(f'Overwrite output fileset: {overwriteOutputFileSet}')
        if overwriteOutputFileSet:
            if os.path.isdir(outputFileSetPath):
                shutil.rmtree(outputFileSetPath)
        os.makedirs(outputFileSetPath, exist_ok=True)
        LOGGER.info(f'Output fileset path: {outputFileSetPath}')

        # Get mode (ARGMAX or PROBABILITIES)
        modeText = self.parameter('mode').value()
        mode = MuscleFatSegmentationTask.ARGMAX if modeText == 'ARGMAX' else MuscleFatSegmentationTask.PROBABILITIES
        LOGGER.info(f'Mode: {modeText}')

        # Load TensorFlow model files
        model, contourModel, parameters = self.loadModelFiles(files=tensorFlowModelFileSet.files())
        if model and parameters:

            # Start iterating of the files
            step = 0
            files = inputFileSet.files()
            segmentationFiles = []
            nrSteps = len(files)
            for file in files:
                try:
                    # Read DICOM file (from cache first) and decompress if needed
                    content = readFromCache(file=file)
                    if not content:
                        p = pydicom.dcmread(file.path())
                        p.decompress()
                        content = writeToCache(file, p)
                    p = content.fileObject()

                    # Get pixels from DICOM file and normalize to positive range
                    img1 = getPixelsFromDicomObject(p, normalize=True)

                    # If contour model provided, apply it to detect abdominal contour
                    if contourModel:
                        mask = self.predictContour(contourModel=contourModel, sourceImage=img1, parameters=parameters)
                        img1 = normalizeBetween(img=img1, minBound=parameters['min_bound'], maxBound=parameters['max_bound'])
                        img1 = img1 * mask
                    else:
                        img1 = normalizeBetween(img=img1, minBound=parameters['min_bound'], maxBound=parameters['max_bound'])

                    img1 = img1.astype(np.float32)
                    img2 = np.expand_dims(img1, 0)
                    img2 = np.expand_dims(img2, -1)
                    pred = model.predict([img2]) ##### Move to separate AI class!
                    predSqueeze = np.squeeze(pred)

                    # Generate predicted output. Can be ARGMAX (pixel value with maximum probability) or
                    # PROBABILITIES, i.e, the individual class probabilities (muscle, SAT and VAT) in 
                    # each pixel
                    if mode == MuscleFatSegmentationTask.ARGMAX:
                        predMax = predSqueeze.argmax(axis=-1)
                        predMax = convertLabelsTo157(labelImage=predMax)
                        segmentationFile = os.path.join(outputFileSetPath, f'{file.name()}.seg.npy')
                        segmentationFiles.append(segmentationFile)
                        np.save(segmentationFile, predMax)
                    elif mode == MuscleFatSegmentationTask.PROBABILITIES:
                        segmentationFile = os.path.join(outputFileSetPath, f'{file.name()}.seg.prob.npy')
                        segmentationFiles.append(segmentationFile)
                        np.save(segmentationFile, predSqueeze)

                except pydicom.errors.InvalidDicomError:
                    LOGGER.warning(f'Skipping non-DICOM: {file.path()}')

                # Update progress for this iteration         
                self.updateProgress(step=step, nrSteps=nrSteps)
                step += 1

        # Build output fileset
        outputFileSet = self.dataManager().createFileSet(fileSetPath=outputFileSetPath)

        # Cache all files in the output fileset because it's very likely we'll be using these
        # files again for further processing and visualization
        for file in outputFileSet.files():
            if file.name().endswith('.seg.npy') or file.name().endswith('.seg.prob.npy'):
                writeToCache(file=file, fileObject=np.load(file.path()))

        LOGGER.info('Finished')


import os

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QApplication, QMainWindow, QWidget, QFileDialog, QMenu, QProgressBar
from PySide6.QtGui import QAction, QGuiApplication

from mosamaticdesktop.data.datamanager import DataManager
from mosamaticdesktop.widgets.datadockwidget import DataDockWidget
from mosamaticdesktop.widgets.taskdockwidget import TaskDockWidget
from mosamaticdesktop.widgets.mainviewerdockwidget import MainViewerDockWidget
from mosamaticdesktop.utils import Configuration

WINDOWTITLE = 'Mosamatic Desktop'
# FILESETPATH = os.path.join(os.getenv('HOME'), 'Desktop', 'downloads', 'dataset', 'scan1')
# FILEPATH = os.path.join(os.getenv('HOME'), 'Desktop', 'downloads', 'dataset', 'scan1', 'image-00000.dcm')


class MainWindow(QMainWindow):
    def __init__(self, version: str, gitHubCommitId: str) -> None:
        super(MainWindow, self).__init__()
        self._version = version
        self._gitHubCommitId = gitHubCommitId
        self._settings = Configuration().qSettings()
        self._dataDockWidget = None
        self._tasksDockWidget = None
        self._mainViewDockWidget = None
        self._defaultLayout = None       
        self._dataManager = DataManager()
        self._dataManager.signal().updated.connect(self.dataUpdated) 
        self.initUi()

    # Initialization

    def initUi(self) -> None:
        self.initActionsAndMenus()
        self.initDataDockWidget()
        self.initTaskDockWidget()
        self.initMainViewerDockWidget()
        self.initMainWindow()

    def initActionsAndMenus(self) -> None:
        importFileAction = QAction('Import File...', self)
        importFileSetAction = QAction('Import File Set...', self)
        deleteAllFileSetsAction = QAction('Delete All Data from Database', self)
        # showApplicationInfoAction = QAction('Show Application Info...', self)
        exitApplicationAction = QAction('Exit', self)
        importFileAction.triggered.connect(self.importFile)
        importFileSetAction.triggered.connect(self.importFileSet)
        deleteAllFileSetsAction.triggered.connect(self.deleteAllFileSets)
        exitApplicationAction.triggered.connect(self.exitApplication)
        dataMenu = QMenu('Data')
        dataMenu.addAction(importFileAction)
        dataMenu.addAction(importFileSetAction)
        dataMenu.addSeparator()
        dataMenu.addAction(deleteAllFileSetsAction)
        dataMenu.addSeparator()
        dataMenu.addAction(exitApplicationAction)
        # aboutMenu = QMenu(f'About')
        # aboutMenu.addAction(showApplicationInfoAction)
        self.menuBar().addMenu(dataMenu)
        # self.menuBar().addMenu(aboutMenu)
        self.menuBar().setNativeMenuBar(False)

    def initDataDockWidget(self) -> None:
        self._dataDockWidget = DataDockWidget(title='Data')
        self._dataDockWidget.setAllowedAreas(Qt.LeftDockWidgetArea | Qt.TopDockWidgetArea)
        self.addDockWidget(Qt.LeftDockWidgetArea, self._dataDockWidget)

    def initTaskDockWidget(self) -> None:
        self._tasksDockWidget = TaskDockWidget(title='Tasks')
        self._tasksDockWidget.setAllowedAreas(Qt.LeftDockWidgetArea | Qt.BottomDockWidgetArea)
        self._tasksDockWidget.setFixedHeight(200)
        self.addDockWidget(Qt.LeftDockWidgetArea, self._tasksDockWidget)

    def initMainViewerDockWidget(self) -> None:
        self._mainViewDockWidget = MainViewerDockWidget(title='Main View')
        self._mainViewDockWidget.setAllowedAreas(Qt.RightDockWidgetArea | Qt.TopDockWidgetArea)
        self.addDockWidget(Qt.RightDockWidgetArea, self._mainViewDockWidget)

    def initMainWindow(self) -> None:
        self.setCentralWidget(QWidget(self))
        self.centralWidget().hide()
        self.splitDockWidget(self._dataDockWidget, self._tasksDockWidget, Qt.Vertical)
        self.setWindowTitle(f'{WINDOWTITLE} {self._version}')
        self.centerWindow()
        self.restoreGeometry(self._settings.value('windowGeometry'))
        self.restoreState(self._settings.value('windowState'))

    # Event handlers

    def importFile(self) -> None:
        lastDirectoryOpened = self._settings.value('lastDirectoryOpened')
        filePath, _ = QFileDialog.getOpenFileName(self, 'Open File', lastDirectoryOpened)
        if filePath:
            self._settings.setValue('lastDirectoryOpened', os.path.split(filePath)[0])
            self._dataManager.createFile(filePath=filePath)

    def importFileSet(self) -> None:
        lastDirectoryOpened = self._settings.value('lastDirectoryOpened')
        fileSetPath = QFileDialog.getExistingDirectory(self, 'Open File Set', lastDirectoryOpened)
        if fileSetPath:
            self._settings.setValue('lastDirectoryOpened', fileSetPath)
            fs = self._dataManager.createFileSet(fileSetPath=fileSetPath)

    def deleteAllFileSets(self) -> None:
        self._dataManager.deleteAllFileSets()

    def dataUpdated(self, value) -> None:
        self._dataDockWidget.treeView().loadFileSetsFromDatabase()

    # Layout and positioning

    def centerWindow(self) -> None:
        screen = QGuiApplication.primaryScreen().geometry()
        x = (screen.width() - self.geometry().width()) / 2
        y = (screen.height() - self.geometry().height()) / 2
        self.move(int(x), int(y))

    # Exit
        
    def exitApplication(self) -> None:
        self._settings.setValue('windowGeometry', self.saveGeometry())
        self._settings.setValue('windowState', self.saveState())
        QApplication.exit()


import os
import shutil

from mosamaticdesktop.tasks.task import Task
from mosamaticdesktop.logger import Logger

LOGGER = Logger()


class CopyFileSetTask(Task):
    def __init__(self) -> None:
        super(CopyFileSetTask, self).__init__()
        self.addDescriptionParameter(
            name='description',
            description='Copies one or more filesets to another fileset'
        )
        self.addMultiFileSetParameter(
            name='inputFileSetNames',
            labelText='Input File Set Names',
        )
        self.addPathParameter(
            name='outputFileSetPath',
            labelText='Output File Set Path',
        )
        self.addTextParameter(
            name='outputFileSetName',
            labelText='Output File Set Name',
            optional=True,
        )
        self.addBooleanParameter(
            name='overwriteOutputFileSet',
            labelText='Overwrite Output File Set',
            defaultValue=True,
        )

    def execute(self) -> None:

        # Get parameters needed for this task
        inputFileSets = []
        inputFileSetNames = self.parameter('inputFileSetNames').value()
        for inputFileSetName in inputFileSetNames:
            inputFileSet = self.dataManager().fileSetByName(name=inputFileSetName)
            inputFileSets.append(inputFileSet)
        outputFileSetPath = self.parameter('outputFileSetPath').value()
        outputFileSetName = self.parameter('outputFileSetName').value()
        if outputFileSetName is None:
            outputFileSetName = self.generateTimestampForFileSetName(name=inputFileSetName)
        overwriteOutputFileSet = self.parameter('overwriteOutputFileSet').value()
        outputFileSetPath = os.path.join(outputFileSetPath, outputFileSetName)
        if overwriteOutputFileSet:
            if os.path.isdir(outputFileSetPath):
                shutil.rmtree(outputFileSetPath)
        os.makedirs(outputFileSetPath, exist_ok=True)
        
        step = 0
        nrSteps = 0
        for inputFileSet in inputFileSets:
            nrSteps += len(inputFileSet.files())
        for inputFileSet in inputFileSets:
            for file in inputFileSet.files():
                shutil.copy(file.path(), outputFileSetPath)
                self.updateProgress(step=step, nrSteps=nrSteps)
                step += 1

        self.dataManager().createFileSet(fileSetPath=outputFileSetPath)
        LOGGER.info('Finished')



from typing import Any

from mosamaticdesktop.data.file import File


class FileContent:
    def __init__(self, file: File, fileObject: Any) -> None:
        self._file = file
        self._fileObject = fileObject

    def file(self) -> File:
        return self._file

    def fileObject(self) -> Any:
        return self._fileObject
    
    def id(self) -> str:
        return self._file.id()


from mosamaticdesktop.tasks.taskwidget import TaskWidget
from mosamaticdesktop.tasks.createarchivetask.createarchivetask import CreateArchiveTask


class CreateArchiveTaskWidget(TaskWidget):
    def __init__(self) -> None:
        super(CreateArchiveTaskWidget, self).__init__(taskType=CreateArchiveTask)
    
    def validate(self) -> None:
        pass


import os
import shutil
import pydicom
import pydicom.errors

from mosamaticdesktop.tasks.task import Task
from mosamaticdesktop.logger import Logger

LOGGER = Logger()


class CheckDicomHeaderTask(Task):
    def __init__(self) -> None:
        super(CheckDicomHeaderTask, self).__init__()
        self.addDescriptionParameter(
            name='description',
            description='Check DICOM headers',
        )
        self.addFileSetParameter(
            name='inputFileSetName',
            labelText='Input File Set',
        )
        self.addTextParameter(
            name='extensionsToIgnore',
            labelText='Extensions to Ignore (comma-separated)',
            defaultValue='tag'
        )
        self.addTextParameter(
            name='requiredAttributes',
            labelText='Required DICOM Attributes',
            defaultValue='Rows, Columns, PixelSpacing',
        )
        self.addIntegerParameter(
            name='rows',
            labelText='Rows',
            defaultValue=512,
            minimum=0,
            maximum=1024,
        )
        self.addIntegerParameter(
            name='columns',
            labelText='Columns',
            defaultValue=512,
            minimum=0,
            maximum=1024,
        )
        self.addPathParameter(
            name='outputFileSetPath',
            labelText='Output File Set Path',
        )
        self.addTextParameter(
            name='outputFileSetName',
            labelText='Output File Set Name',
            optional=True,
        )
        self.addBooleanParameter(
            name='overwriteOutputFileSet',
            labelText='Overwrite Output File Set',
            defaultValue=True,
        )

    def execute(self) -> None:
        inputFileSetName = self.parameter('inputFileSetName').value()
        inputFileSet = self.dataManager().fileSetByName(inputFileSetName)
        requiredAttributes = [x.strip() for x in self.parameter('requiredAttributes').value().split(',')]
        rows = self.parameter('rows').value()
        columns = self.parameter('columns').value()
        outputFileSetPath = self.parameter('outputFileSetPath').value()
        outputFileSetName = self.parameter('outputFileSetName').value()
        if outputFileSetName is None:
            outputFileSetName = self.generateTimestampForFileSetName(name=inputFileSetName)
        overwriteOutputFileSet = self.parameter('overwriteOutputFileSet').value()
        outputFileSetPath = os.path.join(outputFileSetPath, outputFileSetName)
        if overwriteOutputFileSet:
            if os.path.isdir(outputFileSetPath):
                shutil.rmtree(outputFileSetPath)
        os.makedirs(outputFileSetPath, exist_ok=True)
        step = 0
        dicomFilesOk = []
        files = inputFileSet.files()
        nrSteps = len(files)
        for file in files:
            try:
                p = pydicom.dcmread(file.path(), stop_before_pixels=True)
                allAttributesOk = True
                for attribute in requiredAttributes:
                    if attribute not in p:
                        LOGGER.warning(f'{file.path()}: attribute "{attribute}" not present')
                        allAttributesOk = False
                        break
                if allAttributesOk:
                    rowsAndColumnsOk = True
                    if p.Rows != rows:
                        LOGGER.warning(f'{file.path()}: nr. rows should be {rows} but is {p.Rows}')
                        rowsAndColumnsOk = False
                    if p.Columns != columns:
                        LOGGER.warning(f'{file.path()}: Nr. columns should be {columns} but is {p.Columns}')
                        rowsAndColumnsOk = False
                    if rowsAndColumnsOk:
                        dicomFilesOk.append(file)
                    else:
                        pass
                else:
                    pass
            except pydicom.errors.InvalidDicomError:
                LOGGER.warning(f'{file.path()} is not a valid DICOM file')
            self.updateProgress(step=step, nrSteps=nrSteps)
            step += 1
        LOGGER.info(f'Building output fileset...')
        for dicomFile in dicomFilesOk:
            shutil.copy(dicomFile.path(), outputFileSetPath)        
        self.dataManager().createFileSet(fileSetPath=outputFileSetPath)



import os
import json
import shutil
import pydicom
import pydicom.errors
import pandas as pd
import numpy as np

from typing import List, Dict

from mosamaticdesktop.tasks.task import Task
from mosamaticdesktop.data.file import File
from mosamaticdesktop.utils import getPixelsFromDicomObject, tagPixels, isDicomFile, calculateArea, calculateIndex, calculateDiceScore
from mosamaticdesktop.utils import calculateMeanRadiationAttennuation, createNameWithTimestamp, readFromCache, writeToCache
from mosamaticdesktop.logger import Logger

LOGGER = Logger()


class CalculateBodyCompositionMetricsTaskTask(Task):
    MUSCLE = 1
    VAT = 5
    SAT = 7

    def __init__(self) -> None:
        super(CalculateBodyCompositionMetricsTaskTask, self).__init__()
        self.addDescriptionParameter(
            name='description',
            description='Calculates body composition metrics on predicted segmenations and TAG files if available'
        )
        self.addFileSetParameter(
            name='inputFileSetName',
            labelText='Input DICOM (and TAG) File Set',
        )
        self.addFileSetParameter(
            name='inputSegmentationFileSetName',
            labelText='Input Segmentation File Set',
        )
        self.addFilePathParameter(
            name='patientHeightsCsvFilePath',
            labelText='Patient Height CSV File Path (File Name, Height (m))',
            optional=True,
        )
        self.addPathParameter(
            name='outputFileSetPath',
            labelText='Output File Set Path',
        )
        self.addTextParameter(
            name='outputFileSetName',
            labelText='Output File Set Name',
            optional=True,
        )
        self.addBooleanParameter(
            name='overwriteOutputFileSet',
            labelText='Overwrite Output File Set',
            defaultValue=True,
        )

    def findSegmentationFileForDicomFile(self, dicomFile: File, segmentationFiles: List[File]) -> File:
        for segmentationFile in segmentationFiles:
            segmentationFilePath = segmentationFile.path()
            segmentationFileName = os.path.split(segmentationFilePath)[1]
            if dicomFile.name() + '.seg.npy' == segmentationFileName:
                return segmentationFile
        return None

    def findTagFileForDicomFile(self, dicomFile, files: List[File]) -> File:
        tagFilePath1 = dicomFile.path() + '.tag'
        tagFilePath2 = dicomFile.path()[:-4] + '.tag'
        for file in files:            
            if file.path() == tagFilePath1:
                return file
            if file.path() == tagFilePath2:
                return file    
        return None

    def loadDicomFile(self, file: File):
        content = readFromCache(file=file)
        if not content:
            p = pydicom.dcmread(file.path())
            p.decompress()
            content = writeToCache(file, p)
        p = content.fileObject()
        pixelSpacing = p.PixelSpacing
        pixels = getPixelsFromDicomObject(p, normalize=True)
        return pixels, pixelSpacing

    def loadSegmentationFile(self, file: File):
        content = readFromCache(file=file)
        if not content:
            labels = np.load(file.path())
            content = writeToCache(file, labels)
        labels = content.fileObject()
        return labels
    
    def loadTagFile(self, file: File, shape: List[int]):
        content = readFromCache(file=file)
        if not content:
            labels = tagPixels(tagFilePath=file.path())
            content = writeToCache(file, labels)
        labels = content.fileObject()
        return labels.reshape(shape)
    
    def loadPatientHeights(self, df: pd.DataFrame) -> Dict[str, float]:
        data = {}
        for _, row in df.iterrows():
            data[row['file']] = float(row['height'])
        return data
    
    def fileOutputMetricsToString(self, outputMetrics: Dict[str, float]) -> str:
        text = ''
        for metric, value in outputMetrics.items():
            text += f'  - {metric}: {value}\n'
        return text

    def execute(self) -> None:
        # Get input parameters
        inputFileSetName = self.parameter('inputFileSetName').value()
        inputFileSet = self.dataManager().fileSetByName(name=inputFileSetName)
        inputSegmentationFileSetName = self.parameter('inputSegmentationFileSetName').value()
        inputSegmentationFileSet = self.dataManager().fileSetByName(inputSegmentationFileSetName)
        patientHeightsCsvFilePath = self.parameter('patientHeightsCsvFilePath').value()
        patientHeights = None
        if patientHeightsCsvFilePath and patientHeightsCsvFilePath != '':
            patientHeightsDataFrame = pd.read_csv(patientHeightsCsvFilePath, sep='[,;]+', dtype=str)
            patientHeights = self.loadPatientHeights(df=patientHeightsDataFrame)
            LOGGER.info(json.dumps(patientHeights, indent=4))
        outputFileSetName = self.parameter('outputFileSetName').value()
        if not outputFileSetName:
            outputFileSetName = self.generateTimestampForFileSetName(name=inputFileSetName)
        outputFileSetPath = self.parameter('outputFileSetPath').value()
        outputFileSetPath = os.path.join(outputFileSetPath, outputFileSetName)
        overwriteOutputFileSet = self.parameter('overwriteOutputFileSet').value()
        if overwriteOutputFileSet:
            if os.path.isdir(outputFileSetPath):
                shutil.rmtree(outputFileSetPath)
        os.makedirs(outputFileSetPath, exist_ok=True)

        files = inputFileSet.files()
        nrSteps = len(files)
        step = 0
        fileTuples = []
        outputMetrics = {} # Contains BC metrics
        for file in files:
            fileTuple = [None, None, None]

            if isDicomFile(filePath=file.path()):
                fileTuple[0] = file
                tagFile = self.findTagFileForDicomFile(dicomFile=file, files=files)
                if tagFile:
                    fileTuple[2] = tagFile

                # Get segmentation file for DICOM file
                segmentationFile = self.findSegmentationFileForDicomFile(dicomFile=file, segmentationFiles=inputSegmentationFileSet.files())
                if segmentationFile:
                    fileTuple[1] = segmentationFile
                    fileTuples.append(fileTuple)

                    # Get DICOM pixels, pixel spacing, predicted segmentation labels and TAG labels if available
                    image, pixelSpacing = self.loadDicomFile(file=fileTuple[0])
                    segmentation = self.loadSegmentationFile(file=fileTuple[1])
                    
                    # Calculate metrics for predicted segmentation
                    outputMetrics[fileTuple[0]] = {}
                    outputMetrics[fileTuple[0]]['file'] = fileTuple[0].name()
                    outputMetrics[fileTuple[0]]['muscle_area_pred'] = calculateArea(segmentation, CalculateBodyCompositionMetricsTaskTask.MUSCLE, pixelSpacing)
                    if patientHeights:
                        if fileTuple[0].name() in patientHeights.keys():  
                            outputMetrics[fileTuple[0]]['muscle_index_pred'] = calculateIndex(
                                area=outputMetrics[fileTuple[0]]['muscle_area_pred'], height=patientHeights[fileTuple[0].name()])
                        else:
                            outputMetrics[fileTuple[0]]['muscle_index_pred'] = 0
                    outputMetrics[fileTuple[0]]['vat_area_pred'] = calculateArea(segmentation, CalculateBodyCompositionMetricsTaskTask.VAT, pixelSpacing)
                    if patientHeights:
                        if fileTuple[0].name() in patientHeights.keys():  
                            outputMetrics[fileTuple[0]]['vat_index_pred'] = calculateIndex(
                                area=outputMetrics[fileTuple[0]]['vat_area_pred'], height=patientHeights[fileTuple[0].name()])
                        else:
                            outputMetrics[fileTuple[0]]['vat_index_pred'] = 0
                    outputMetrics[fileTuple[0]]['sat_area_pred'] = calculateArea(segmentation, CalculateBodyCompositionMetricsTaskTask.SAT, pixelSpacing)
                    if patientHeights:
                        if fileTuple[0].name() in patientHeights.keys():
                            outputMetrics[fileTuple[0]]['sat_index_pred'] = calculateIndex(
                                area=outputMetrics[fileTuple[0]]['sat_area_pred'], height=patientHeights[fileTuple[0].name()])
                        else:
                            outputMetrics[fileTuple[0]]['sat_index_pred'] = 0
                    outputMetrics[fileTuple[0]]['muscle_ra_pred'] = calculateMeanRadiationAttennuation(image, segmentation, CalculateBodyCompositionMetricsTaskTask.MUSCLE)
                    outputMetrics[fileTuple[0]]['vat_ra_pred'] = calculateMeanRadiationAttennuation(image, segmentation, CalculateBodyCompositionMetricsTaskTask.VAT)
                    outputMetrics[fileTuple[0]]['sat_ra_pred'] = calculateMeanRadiationAttennuation(image, segmentation, CalculateBodyCompositionMetricsTaskTask.SAT)

                    if tagFile:
                        # Calculate metrics for true segmentation based on TAG file
                        tagImage = self.loadTagFile(file=fileTuple[2], shape=image.shape)
                        outputMetrics[fileTuple[0]]['muscle_area_true'] = calculateArea(tagImage, CalculateBodyCompositionMetricsTaskTask.MUSCLE, pixelSpacing)
                        if patientHeights:
                            outputMetrics[fileTuple[0]]['muscle_index_true'] = calculateIndex(
                                area=outputMetrics[fileTuple[0]]['muscle_area_true'], height=patientHeights[fileTuple[0].name()])
                        outputMetrics[fileTuple[0]]['vat_area_true'] = calculateArea(tagImage, CalculateBodyCompositionMetricsTaskTask.VAT, pixelSpacing)
                        if patientHeights:
                            outputMetrics[fileTuple[0]]['vat_index_true'] = calculateIndex(
                                area=outputMetrics[fileTuple[0]]['vat_area_true'], height=patientHeights[fileTuple[0].name()])
                        outputMetrics[fileTuple[0]]['sat_area_true'] = calculateArea(tagImage, CalculateBodyCompositionMetricsTaskTask.SAT, pixelSpacing)
                        if patientHeights:
                            outputMetrics[fileTuple[0]]['sat_index_true'] = calculateIndex(
                                area=outputMetrics[fileTuple[0]]['sat_area_true'], height=patientHeights[fileTuple[0].name()])
                        outputMetrics[fileTuple[0]]['muscle_ra_true'] = calculateMeanRadiationAttennuation(image, tagImage, CalculateBodyCompositionMetricsTaskTask.MUSCLE)
                        outputMetrics[fileTuple[0]]['vat_ra_true'] = calculateMeanRadiationAttennuation(image, tagImage, CalculateBodyCompositionMetricsTaskTask.VAT)
                        outputMetrics[fileTuple[0]]['sat_ra_true'] = calculateMeanRadiationAttennuation(image, tagImage, CalculateBodyCompositionMetricsTaskTask.SAT)

                        # Calculate Dice score between segmentation and TAG files
                        outputMetrics[fileTuple[0]]['dice_muscle'] = calculateDiceScore(groundTruth=tagImage, prediction=segmentation, label=CalculateBodyCompositionMetricsTaskTask.MUSCLE)
                        outputMetrics[fileTuple[0]]['dice_vat'] = calculateDiceScore(groundTruth=tagImage, prediction=segmentation, label=CalculateBodyCompositionMetricsTaskTask.VAT)
                        outputMetrics[fileTuple[0]]['dice_sat'] = calculateDiceScore(groundTruth=tagImage, prediction=segmentation, label=CalculateBodyCompositionMetricsTaskTask.SAT)
                    LOGGER.info(self.fileOutputMetricsToString(outputMetrics=outputMetrics[fileTuple[0]]))
                else:
                    LOGGER.error(f'Segmentation file {segmentationFile.name()} not found')
        
            # Update progress
            self.updateProgress(step=step, nrSteps=nrSteps)
            step += 1
        
        # Build output fileset using Pandas
        LOGGER.info(f'Building output fileset: {outputFileSetPath}...')
        firstKey = next(iter(outputMetrics))
        columns = list(outputMetrics[firstKey].keys())
        data = {}
        for column in columns:
            data[column] = []
        for filePath in outputMetrics.keys():
            for column in columns:
                data[column].append(outputMetrics[filePath][column])
        csvFilePath = os.path.join(outputFileSetPath, createNameWithTimestamp('scores') + '.csv')
        df = pd.DataFrame(data=data)
        df.to_csv(csvFilePath, index=False)

        # Create new output fileset from the CSV file. No need to cache it because it's just
        # text and we're probably not going to use it further
        self.dataManager().createFileSet(fileSetPath=outputFileSetPath)
        LOGGER.info('Finished')


from typing import Any, List

from PySide6.QtWidgets import QWidget, QLabel, QComboBox

from mosamaticdesktop.tasks.optiongroupparameter import OptionGroupParameter
from mosamaticdesktop.tasks.parameterwidget import ParameterWidget


class OptionGroupParameterWidget(ParameterWidget):
    def __init__(self, parameter: OptionGroupParameter, parent: QWidget=None) -> None:
        super(OptionGroupParameterWidget, self).__init__(parameter=parameter, parent=parent)
        self._optionGroupComboBox = None
        self.initUi()

    def initUi(self) -> None:
        self._optionGroupComboBox = QComboBox(self)
        self._optionGroupComboBox.addItem(None)
        for option in self.parameter().options():
            self._optionGroupComboBox.addItem(option)
        if self.parameter().defaultValue() is not None:
            self._optionGroupComboBox.setCurrentText(self.parameter().defaultValue())
            self.parameter().setValue(self.parameter().defaultValue())
        self._optionGroupComboBox.currentTextChanged.connect(self.currentTextChanged)
        self.layout().addWidget(QLabel(self.parameter().labelText()))
        self.layout().addWidget(self._optionGroupComboBox)

    def currentTextChanged(self, text: str) -> None:
        self.parameter().setValue(text)


from mosamaticdesktop.data.models.filemodel import FileModel


class File:
    def __init__(self, model: FileModel) -> None:
        self._id = model.id
        self._name = model.name
        self._path = model.path

    def id(self) -> str:
        return self._id

    def name(self) -> str:
        return self._name

    def path(self) -> str:
        return self._path


from typing import Any

from PySide6.QtWidgets import QWidget, QVBoxLayout, QLabel, QVBoxLayout, QDoubleSpinBox

from mosamaticdesktop.tasks.parameter import Parameter
from mosamaticdesktop.tasks.parameterwidget import ParameterWidget


class FloatingPointParameterWidget(ParameterWidget):
    def __init__(self, parameter: Parameter, parent: QWidget=None) -> None:
        super(FloatingPointParameterWidget, self).__init__(parameter=parameter, parent=parent)
        if self.parameter().defaultValue() is not None:
            self.parameter().setValue(self.parameter().defaultValue())
        self._floatingPointSpinBox = None
        self.initUi()

    def initUi(self) -> None:
        self._floatingPointSpinBox = QDoubleSpinBox(self)
        self._floatingPointSpinBox.setValue(self.parameter().defaultValue())
        self._floatingPointSpinBox.valueChanged.connect(self.valueChanged)
        self.layout().addWidget(QLabel(self.parameter().labelText()))
        self.layout().addWidget(self._floatingPointSpinBox)

    def valueChanged(self, value: int) -> None:
        self.parameter().setValue(value)


from typing import Any, List

from PySide6.QtWidgets import QWidget, QHBoxLayout, QLabel, QComboBox, QListWidget, QPushButton

from mosamaticdesktop.tasks.parameter import Parameter
from mosamaticdesktop.tasks.parameterwidget import ParameterWidget
from mosamaticdesktop.data.datamanager import DataManager


class MultiFileSetParameterWidget(ParameterWidget):
    class ComboBox(QComboBox):
        def __init__(self, parent: QWidget=None) -> None:
            super(MultiFileSetParameterWidget.ComboBox, self).__init__(parent=parent)
            self._dataManager = DataManager()
            self.loadItems()

        def loadItems(self) -> None:
            self.addItem(None)
            fileSets = self._dataManager.fileSets()
            for fileSet in fileSets:
                self.addItem(fileSet.name())

    def __init__(self, parameter: Parameter, parent: QWidget=None) -> None:
        super(MultiFileSetParameterWidget, self).__init__(parameter=parameter, parent=parent)
        self._fileSetComboBox = None
        self._fileSetItemListWidget = None
        self.initUi()

    def initUi(self) -> None:
        label = QLabel(self.parameter().labelText())
        self._fileSetComboBox = MultiFileSetParameterWidget.ComboBox(self)
        addButton = QPushButton('Add', self)
        addButton.clicked.connect(self.addFileSet)
        removeButton = QPushButton('Remove', self)
        removeButton.clicked.connect(self.removeFileSet)
        buttonLayout = QHBoxLayout()
        buttonLayout.addWidget(addButton)
        buttonLayout.addWidget(removeButton)
        self._fileSetItemListWidget = QListWidget(self)
        self.layout().addWidget(label)
        self.layout().addWidget(self._fileSetComboBox)
        self.layout().addLayout(buttonLayout)
        self.layout().addWidget(self._fileSetItemListWidget)

    def addFileSet(self):
        itemNameIndex = self._fileSetComboBox.currentIndex()
        itemName = self._fileSetComboBox.itemText(itemNameIndex)
        if itemName and itemName != '':
            self._fileSetItemListWidget.addItem(itemName)
            self._fileSetComboBox.removeItem(itemNameIndex)
            self.parameter().setValue(self.items())

    def removeFileSet(self):
        item = self._fileSetItemListWidget.currentItem()
        if item:
            self._fileSetItemListWidget.takeItem(self._fileSetItemListWidget.row(item))
            self._fileSetComboBox.addItem(item.text())
            self.parameter().setValue(self.items())

    def items(self) -> List[str]:
        items = []
        for i in range(self._fileSetItemListWidget.count()):
            items.append(self._fileSetItemListWidget.item(i).text())
        return items


from typing import Any

from PySide6.QtWidgets import QWidget, QLabel, QSizePolicy
from PySide6.QtGui import QFont

from mosamaticdesktop.tasks.parameter import Parameter
from mosamaticdesktop.tasks.parameterwidget import ParameterWidget


class LabelParameterWidget(ParameterWidget):
    def __init__(self, parameter: Parameter, parent: QWidget=None) -> None:
        super(LabelParameterWidget, self).__init__(parameter=parameter, parent=parent)
        self.initUi()

    def initUi(self) -> None:
        font = QFont()
        font.setItalic(True)
        label = QLabel('')
        label.setFont(font)
        label.setWordWrap(True)
        label.setText(self.parameter().labelText())
        policy = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        label.setSizePolicy(policy)
        self.layout().addWidget(label)


from typing import Any

from mosamaticdesktop.tasks.parameter import Parameter


class LabelParameter(Parameter):
    def __init__(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None) -> None:
        super(LabelParameter, self).__init__(
            name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue)

    def copy(self):
        return LabelParameter(
            name=self.name(), 
            labelText=self.labelText(), 
            optional=self.optional(), 
            visible=self.visible(), 
            defaultValue=self.defaultValue(),
        )


from mosamaticdesktop.tasks.taskwidget import TaskWidget
from mosamaticdesktop.tasks.calculatebodycompositionmetricstask.calculatebodycompositionmetricstask import CalculateBodyCompositionMetricsTaskTask


class CalculateBodyCompositionMetricsTaskWidget(TaskWidget):
    def __init__(self) -> None:
        super(CalculateBodyCompositionMetricsTaskWidget, self).__init__(taskType=CalculateBodyCompositionMetricsTaskTask)
    
    def validate(self) -> None:
        pass


from typing import Any

from mosamaticdesktop.tasks.parameter import Parameter


class IntegerParameter(Parameter):
    def __init__(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None, minimum: int=0, maximum: int=100, step: int=1) -> None:
        super(IntegerParameter, self).__init__(
            name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue)
        if self.defaultValue() is not None:
            self.setValue(self.defaultValue())
        self._minimum = minimum
        self._maximum = maximum
        self._step = step

    def minimum(self) -> int:
        return self._minimum
    
    def maximum(self) -> int:
        return self._maximum
    
    def step(self) -> int:
        return self._step

    def copy(self):
        return IntegerParameter(
            name=self.name(), 
            labelText=self.labelText(), 
            optional=self.optional(), 
            visible=self.visible(), 
            defaultValue=self.defaultValue(),
            minimum=self.minimum(),
            maximum=self.maximum(),
            step=self.step(),
        )


import os
import time
import sys
import math
import pickle
import pendulum
import importlib
import pydicom
import binascii
import struct
import pydicom.errors
import numpy as np
# import matplotlib
# matplotlib.use('Qt5Agg')
# import matplotlib.pyplot as plt

from PIL import Image
from typing import Dict, Any, List

from PySide6.QtGui import QImage
from PySide6.QtCore import QSettings

from mosamaticdesktop.singleton import singleton
from mosamaticdesktop.operatingsystem import OperatingSystem
from mosamaticdesktop.data.file import File
from mosamaticdesktop.data.filecontent import FileContent
from mosamaticdesktop.data.filecontentcache import FileContentCache


def createNameWithTimestamp(prefix: str='') -> str:
    tz = pendulum.local_timezone()
    timestamp = pendulum.now(tz).strftime('%Y%m%d%H%M%S%f')[:17]
    if prefix != '' and not prefix.endswith('-'):
        prefix = prefix + '-'
    name = f'{prefix}{timestamp}'
    return name


def currentTimeInMilliseconds() -> int:
    return int(round(time.time() * 1000))


def currentTimeInSeconds() -> int:
    return int(round(currentTimeInMilliseconds() / 1000.0))


def elapsedMilliseconds(startTimeInMilliseconds: int) -> int:
    return currentTimeInMilliseconds() - startTimeInMilliseconds


def elapsedSeconds(startTimeInSeconds: int) -> int:
    return currentTimeInSeconds() - startTimeInSeconds


def duration(seconds: int) -> str:
    h = int(math.floor(seconds/3600.0))
    remainder = seconds - h * 3600
    m = int(math.floor(remainder/60.0))
    remainder = remainder - m * 60
    s = int(math.floor(remainder))
    return '{} hours, {} minutes, {} seconds'.format(h, m, s)


def readFromCache(file: File) -> FileContent:
    cache = FileContentCache()
    return cache.get(id=file.id())


def writeToCache(file: File, fileObject: Any) -> None:
    content = FileContent(file=file, fileObject=fileObject)
    cache = FileContentCache()
    cache.add(content)
    # Return content rightaway for continued processing
    # after updating cache
    return content


def isDicomFile(filePath: str) -> bool:
    try:
        pydicom.dcmread(filePath, stop_before_pixels=True)
        return True
    except pydicom.errors.InvalidDicomError:
        return False
    

def isNumPyFile(filePath: str) -> bool:
    return filePath.endswith('.npy')
    

def isTagFile(filePath: str) -> bool:
    return filePath.endswith('.tag')


def getPixelsFromDicomObject(p: pydicom.FileDataset, normalize: bool=False) -> np.array:
    pixels = p.pixel_array
    if not normalize:
        return pixels
    if normalize is True:
        return p.RescaleSlope * pixels + p.RescaleIntercept
    if isinstance(normalize, int):
        return (pixels + np.min(pixels)) / (np.max(pixels) - np.min(pixels)) * normalize
    if isinstance(normalize, list):
        return (pixels + np.min(pixels)) / (np.max(pixels) - np.min(pixels)) * normalize[1] + normalize[0]
    return pixels


def convertLabelsTo157(labelImage: np.array) -> np.array:
    labelImage157 = np.copy(labelImage)
    labelImage157[labelImage157 == 1] = 1
    labelImage157[labelImage157 == 2] = 5
    labelImage157[labelImage157 == 3] = 7
    return labelImage157


def normalizeBetween(img: np.array, minBound: int, maxBound: int) -> np.array:
    img = (img - minBound) / (maxBound - minBound)
    img[img > 1] = 0
    img[img < 0] = 0
    c = (img - np.min(img))
    d = (np.max(img) - np.min(img))
    img = np.divide(c, d, np.zeros_like(c), where=d != 0)
    return img


def applyWindowCenterAndWidth(image: np.array, center: int, width: int) -> np.array:
    imageMin = center - width // 2
    imageMax = center + width // 2
    windowedImage = np.clip(image, imageMin, imageMax)
    windowedImage = ((windowedImage - imageMin) / (imageMax - imageMin)) * 255.0
    return windowedImage.astype(np.uint8)


class ColorMap:
    def __init__(self, name: str) -> None:
        self._name = name
        self._values = []

    def name(self) -> str:
        return self._name
    
    def values(self) -> List[List[int]]:
        return self._values
    

class GrayScaleColorMap(ColorMap):
    def __init__(self) -> None:
        super(GrayScaleColorMap, self).__init__(name='GrayScaleColorMap')
        # Implement your own gray scale map or let NumPy do this more efficiently?
        pass    

class AlbertaColorMap(ColorMap):
    def __init__(self) -> None:
        super(AlbertaColorMap, self).__init__(name='AlbertaColorMap')
        for i in range(256):
            if i == 1:  # muscle
                self.values().append([255, 0, 0])
            elif i == 2:  # inter-muscular adipose tissue
                self.values().append([0, 255, 0])
            elif i == 5:  # visceral adipose tissue
                self.values().append([255, 255, 0])
            elif i == 7:  # subcutaneous adipose tissue
                self.values().append([0, 255, 255])
            elif i == 12:  # unknown
                self.values().append([0, 0, 255])
            else:
                self.values().append([0, 0, 0])


def applyColorMap(pixels: np.array, colorMap: ColorMap) -> np.array:
    pixelsNew = np.zeros((*pixels.shape, 3), dtype=np.uint8)
    np.take(colorMap.values(), pixels, axis=0, out=pixelsNew)
    return pixelsNew


def tagPixels(tagFilePath: str) -> np.array:
    f = open(tagFilePath, 'rb')
    f.seek(0)
    byte = f.read(1)
    # Make sure to check the byte-value in Python 3!!
    while byte != b'':
        byteHex = binascii.hexlify(byte)
        if byteHex == b'0c':
            break
        byte = f.read(1)
    values = []
    f.read(1)
    while byte != b'':
        v = struct.unpack('b', byte)
        values.append(v)
        byte = f.read(1)
    values = np.asarray(values)
    values = values.astype(np.uint16)
    return values


def loadNumPyArray(filePath: str) -> np.array:
    return np.load(filePath)


def convertDicomToNumPyArray(dicomFilePath: str, windowLevel: int=50, windowWidth: int=400, normalize=True) -> np.array:
    p = pydicom.dcmread(dicomFilePath)
    pixels = p.pixel_array
    pixels = pixels.reshape(p.Rows, p.Columns)
    if normalize:
        b = p.RescaleIntercept
        m = p.RescaleSlope
        pixels = m * pixels + b
    pixels = applyWindowCenterAndWidth(pixels, windowLevel, windowWidth)
    return pixels


def convertNumPyArrayToPngImage(
        numpyArrayFilePathOrObject: str, outputDirectoryPath: str, colorMap: ColorMap=None, pngImageFileName: str=None, figureWidth: int=10, figureHeight: int=10) -> str:
    if isinstance(numpyArrayFilePathOrObject, str):
        numpyArray = loadNumPyArray(numpyArrayFilePathOrObject)
    else:
        numpyArray = numpyArrayFilePathOrObject
        if not pngImageFileName:
            raise RuntimeError('PNG file name required for NumPy array object')
    if colorMap:
        numpyArray = applyColorMap(pixels=numpyArray, colorMap=colorMap)
    image = Image.fromarray(numpyArray)
    # fig = plt.figure(figsize=(figureWidth, figureHeight))
    # ax = fig.add_subplot(1, 1, 1)
    # if colorMap:
    #     plt.imshow(numpyArray)
    # else:
    #     plt.imshow(numpyArray, cmap='gray')
    # ax.axis('off')
    if not pngImageFileName:
        numpyArrayFileName = os.path.split(numpyArrayFilePathOrObject)[1]
        pngImageFileName = numpyArrayFileName + '.png'      
    elif not pngImageFileName.endswith('.png'):
        pngImageFileName += '.png'
    pngImageFilePath = os.path.join(outputDirectoryPath, pngImageFileName)
    # plt.savefig(pngImageFilePath, bbox_inches='tight')
    # plt.close('all')
    image.save(pngImageFilePath)
    return pngImageFilePath


def convertNumPyArrayToRgbQImage(numpyArray: np.array, colorMap: ColorMap) -> QImage:
    numpyArray = applyColorMap(pixels=numpyArray, colorMap=colorMap)
    h, w, _ = numpyArray.shape
    image = QImage(numpyArray.data.tobytes(), w, h, QImage.Format_RGB888)
    return image


def calculateArea(labels, label, pixelSpacing):
    mask = np.copy(labels)
    mask[mask != label] = 0
    mask[mask == label] = 1
    area = np.sum(mask) * (pixelSpacing[0] * pixelSpacing[1]) / 100.0
    return area


def calculateIndex(area: float, height: float):
    return area / (height * height)


def calculateMeanRadiationAttennuation(image, labels, label):
    mask = np.copy(labels)
    mask[mask != label] = 0
    mask[mask == label] = 1
    subtracted = image * mask
    maskSum = np.sum(mask)
    if maskSum > 0.0:
        meanRadiationAttenuation = np.sum(subtracted) / np.sum(mask)
    else:
        # print('Sum of mask pixels is zero, return zero radiation attenuation')
        meanRadiationAttenuation = 0.0
    return meanRadiationAttenuation


def calculateDiceScore(groundTruth, prediction, label):
    numerator = prediction[groundTruth == label]
    numerator[numerator != label] = 0
    n = groundTruth[prediction == label]
    n[n != label] = 0
    if np.sum(numerator) != np.sum(n):
        raise RuntimeError('Mismatch in Dice score calculation!')
    denominator = (np.sum(prediction[prediction == label]) + np.sum(groundTruth[groundTruth == label]))
    diceScore = np.sum(numerator) * 2.0 / denominator
    return diceScore


@singleton
class GitCommit:
    def __init__(self) -> None:
        commitIdFilePath = os.path.join(os.path.dirname(sys.executable), 'gitcommitid.txt')
        if not os.path.isfile(commitIdFilePath):
            commitIdFilePath = 'gitcommitid.txt'
        self._commitId = open(commitIdFilePath, 'r').readline().strip()

    def id(self) -> str:
        return self._commitId
    

@singleton
class Configuration:
    def __init__(self) -> None:
        self._configDirectory = os.path.join(OperatingSystem.homeDirectory(), '.mosamatic')

    def configDirectory(self) -> str:
        return self._configDirectory
    
    def taskConfigDirectory(self, taskName: str) -> str:
        taskConfigDirectory = os.path.join(self._configDirectory, taskName)
        if not os.path.isdir(taskConfigDirectory):
            os.makedirs(taskConfigDirectory, exist_ok=False)
        return taskConfigDirectory
    
    def taskConfigSubDirectory(self, taskName: str, dirName: str) -> str:
        taskConfigDirectory = self.taskConfigDirectory(taskName=taskName)
        taskConfigSubDirectory = os.path.join(taskConfigDirectory, dirName)
        if not os.path.isdir(taskConfigSubDirectory):
            os.makedirs(taskConfigSubDirectory, exist_ok=False)
        return taskConfigSubDirectory
    
    def qSettings(self) -> QSettings:
        settingsPath = os.path.join(self.configDirectory(), 'settings.ini')
        return QSettings(settingsPath, QSettings.Format.IniFormat)
    

class ModuleLoader:
    @staticmethod
    def loadModuleClasses(moduleDirectoryPath: str, moduleBaseClass: Any, fileNameEndsWith: str) -> Dict[str, Any]:
        classes = {}
        moduleDirectoryName = os.path.split(moduleDirectoryPath)[1]
        for root, dirs, files in os.walk(moduleDirectoryPath):
            for fileName in files:
                filePath = os.path.join(root, fileName)
                if fileName.endswith(fileNameEndsWith):
                    taskModule = filePath.split(os.path.sep)[-2]
                    if taskModule != moduleDirectoryName:
                        spec = importlib.util.spec_from_file_location(taskModule, filePath)
                        if spec and spec.loader:
                            module = importlib.util.module_from_spec(spec)
                            spec.loader.exec_module(module)
                            for attributeName in dir(module):
                                attribute = getattr(module, attributeName)
                                if isinstance(attribute, type) and issubclass(attribute, moduleBaseClass) and attribute is not moduleBaseClass:
                                    classes[attribute.NAME()] = attribute
        return classes
    

class ParameterLoader:
    @staticmethod
    def loadParameterClasses(parameterDirectoryPath: str, parameterBaseClass: Any) -> Dict[str, Any]:
        classes = {}
        parameterDirectoryName = os.path.split(parameterDirectoryPath)[1]
        for root, dirs, files in os.walk(parameterDirectoryPath):
            for fileName in files:
                filePath = os.path.join(root, fileName)
                if fileName.endswith('parameter.py'):
                    taskModule = filePath.split(os.path.sep)[-1]
                    if taskModule != parameterDirectoryName:
                        taskModule = f'tasks.{taskModule[:-3]}'
                        spec = importlib.util.spec_from_file_location(taskModule, filePath)
                        if spec and spec.loader:
                            module = importlib.util.module_from_spec(spec)
                            spec.loader.exec_module(module)
                            for attributeName in dir(module):
                                attribute = getattr(module, attributeName)
                                if isinstance(attribute, type) and issubclass(attribute, parameterBaseClass) and attribute is not parameterBaseClass:
                                    classes[attribute.NAME()] = attribute
        return classes



import uuid

from sqlalchemy import Column, String, ForeignKey
from sqlalchemy.orm import Mapped, relationship

from mosamaticdesktop.data.models.basemodel import BaseModel


class FileModel(BaseModel):
    __tablename__ = '_filemodel'
    id: Mapped[str] = Column('_id', String, primary_key=True, default=lambda: str(uuid.uuid4()), unique=True, nullable=False)
    name: Mapped[str] = Column('_name', String(256), nullable=False)
    path: Mapped[str] = Column('_path', String(1024), nullable=False)
    fileSetModel: Mapped['FileSetModel'] = relationship('FileSetModel', back_populates='fileModels')
    fileSetModelId: Mapped[str] = Column('_filesetmodel_id', ForeignKey('_filesetmodel._id'))



import os
import shutil
import pydicom
import pydicom.errors

from mosamaticdesktop.tasks.task import Task
from mosamaticdesktop.data.filecontentcache import FileContentCache
from mosamaticdesktop.utils import readFromCache, writeToCache
from mosamaticdesktop.logger import Logger

LOGGER = Logger()


class DecompressDicomTask(Task):
    def __init__(self) -> None:
        super(DecompressDicomTask, self).__init__()
        self.addDescriptionParameter(
            name='description',
            description='Decompresses JPEG200-compressed DICOM files for use in Slice-o-matic.'
        )
        self.addFileSetParameter(
            name='inputFileSetName',
            labelText='Input File Set Name',
        )
        self.addPathParameter(
            name='outputFileSetPath',
            labelText='Output File Set Path',
        )
        self.addTextParameter(
            name='outputFileSetName',
            labelText='Output File Set Name',
            optional=True,
        )
        self.addBooleanParameter(
            name='overwriteOutputFileSet',
            labelText='Overwrite Output File Set',
            defaultValue=True,
        )
        self._cache = FileContentCache()

    def execute(self) -> None:
        # Prepare input parameters
        inputFileSetName = self.parameter('inputFileSetName').value()
        inputFileSet = self.dataManager().fileSetByName(inputFileSetName)
        outputFileSetPath = self.parameter('outputFileSetPath').value()
        outputFileSetName = self.parameter('outputFileSetName').value()
        if outputFileSetName is None:
            outputFileSetName = self.generateTimestampForFileSetName(name=inputFileSetName)

        overwriteOutputFileSet = self.parameter('overwriteOutputFileSet').value()
        outputFileSetPath = os.path.join(outputFileSetPath, outputFileSetName)
        
        if overwriteOutputFileSet:
            if os.path.isdir(outputFileSetPath):
                shutil.rmtree(outputFileSetPath)
        os.makedirs(outputFileSetPath, exist_ok=True)

        step = 0
        files = inputFileSet.files()
        nrSteps = len(files)
        for file in files:
            try:
                # Try to load file content from cache first. If it's not available
                # read it from disk
                content = readFromCache(file=file)
                if not content:
                    p = pydicom.dcmread(file.path())
                    p.decompress()
                    content = writeToCache(file, p)
                p = content.fileObject()
                outputFileName = file.name()
                outputFilePath = os.path.join(outputFileSetPath, outputFileName)
                p.save_as(outputFilePath)
            except pydicom.errors.InvalidDicomError:
                LOGGER.warning(f'Skipping non-DICOM: {file.path()}')
            self.updateProgress(step=step, nrSteps=nrSteps)
            step += 1
        self.dataManager().createFileSet(fileSetPath=outputFileSetPath)
        LOGGER.info('Finished')


from PySide6.QtCore import QPoint
from PySide6.QtWidgets import QMenu, QTreeView, QMessageBox

from mosamaticdesktop.widgets.filesetitem import FileSetItem
from mosamaticdesktop.data.datamanager import DataManager


class FileSetItemMenu(QMenu):
    def __init__(self, treeView: QTreeView, fileSetItem: FileSetItem, position: QPoint) -> None:
        super(FileSetItemMenu, self).__init__()
        self._treeView = treeView
        self._fileSetItem = fileSetItem
        self._position = position
        self._dataManager = DataManager()
        renameFileSetAction = self.addAction('Rename')
        renameFileSetAction.triggered.connect(self.renameFileSet)
        deleteFileSetAction = self.addAction('Delete')
        deleteFileSetAction.triggered.connect(self.deleteFileSet)

    def renameFileSet(self):
        self._fileSetItem.setEditable(True)
        self._treeView.edit(self._treeView.model().indexFromItem(self._fileSetItem))
        self._fileSetItem.setEditable(False)

    def deleteFileSet(self):
        fileSet = self._fileSetItem.fileSet()
        self._dataManager.deleteFileSet(fileSet=fileSet)

    def show(self):
        self.exec_(self._position)


from PySide6.QtCore import Qt, QPoint
from PySide6.QtWidgets import QTreeView
from PySide6.QtGui import QResizeEvent, QStandardItemModel, QStandardItem, QMouseEvent

from mosamaticdesktop.data.datamanager import DataManager
from mosamaticdesktop.data.fileset import FileSet
from mosamaticdesktop.widgets.fileitem import FileItem
from mosamaticdesktop.widgets.fileitemmenu import FileItemMenu
from mosamaticdesktop.widgets.filesetitem import FileSetItem
from mosamaticdesktop.widgets.filesetitemmenu import FileSetItemMenu
from mosamaticdesktop.widgets.multifilesetitemmenu import MultiFileSetItemMenu

FIRSTCOLUMNOFFSET = 100


class FileSetTreeView(QTreeView):
    def __init__(self) -> None:
        super(FileSetTreeView, self).__init__()
        self._model = None
        self._dataManager = DataManager()
        self.initModel()
        self.loadFileSetsFromDatabase()

    def initModel(self) -> None:
        self._model = QStandardItemModel()
        self._model.setHorizontalHeaderLabels(['File Sets', 'Nr. Files'])
        self._model.itemChanged.connect(self.itemChanged)
        self.setModel(self._model)
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.setSelectionMode(QTreeView.ExtendedSelection)

    def itemChanged(self, item: QStandardItem) -> None:
        if isinstance(item, FileSetItem):
            fileSet = item.fileSet()
            fileSet.setName(item.text())
            self._dataManager.updateFileSet(fileSet=fileSet)

    def loadFileSetsFromDatabase(self) -> None:
        self.clearFileSets()
        for fileSet in self._dataManager.fileSets():
            self.addFileSet(fileSet=fileSet)

    def addFileSet(self, fileSet: FileSet) -> None:
        fileSetItem = FileSetItem(fileSet=fileSet)
        fileSetItem.setEditable(False)
        for file in fileSet.files():
            fileItem = FileItem(file=file)
            fileItem.setEditable(False)
            fileSetItem.appendRow([fileItem, QStandardItem()])
        self._model.appendRow([fileSetItem, QStandardItem(str(fileSetItem.fileSet().nrFiles()))])

    def mousePressEvent(self, event: QMouseEvent) -> None:
        index = self.indexAt(event.pos())
        if index.isValid():
            globalPos = self.viewport().mapToGlobal(event.pos())
            if event.button() == Qt.RightButton:
                self.rightClickEvent(index, globalPos)
                return
        super(FileSetTreeView, self).mousePressEvent(event)

    def rightClickEvent(self, index: int, globalPos: QPoint) -> None:
        selectedIndexes = self.selectionModel().selectedIndexes()
        selectedItems = [self._model.itemFromIndex(idx) for idx in selectedIndexes]
        if len(selectedItems) == 1:
            item = selectedItems[0]
            if isinstance(item, FileItem):
                menu = FileItemMenu(self, item, globalPos)
                menu.show()
            elif isinstance(item, FileSetItem):
                menu = FileSetItemMenu(self, item, globalPos)
                menu.show()
            else:
                pass
        elif len(selectedItems) > 1:
            item = selectedItems[0]
            if isinstance(item, FileSetItem):
                menu = MultiFileSetItemMenu(self, selectedItems, globalPos)
                menu.show()
        else:
            pass

    def resizeEvent(self, event: QResizeEvent) -> None:
        self.setColumnWidth(0, event.size().width() - FIRSTCOLUMNOFFSET)
        return super().resizeEvent(event)

    def clearFileSets(self) -> None:
        self._model.clear()


import numpy as np


class Model:
    def __init__(self) -> None:
        self._model = None

    def load(self, modelFilePath: str) -> None:
        raise NotImplementedError()

    def predict(self, input: np.array) -> np.array:
        raise NotImplementedError()


import os
import sys

from PySide6.QtWidgets import QApplication

# from mainwindow import MainWindow
from mosamaticdesktop.mainwindow import MainWindow

versionFile = os.path.join(os.path.split(os.path.abspath(__file__))[0], 'VERSION')
with open(versionFile) as f:
    VERSION = f.read()

# gitHubCommitIdFile = os.path.join(os.path.split(os.path.abspath(__file__))[0], 'GIT_COMMIT_ID')
# with open(gitHubCommitIdFile) as f:
#     GIT_COMMIT_ID = f.read()


def main():
    app = QApplication([])
    mainWindow = MainWindow(version=VERSION, gitHubCommitId=None)
    mainWindow.show()
    sys.exit(app.exec())
    pass


if __name__ == "__main__":
    main()


import os
import shutil
from starlette.background import BackgroundTasks
from enum import Enum
import threading
import time
import logging
import secrets
import numpy as np
from totalsegmentator.map_to_binary import class_map
from scipy.ndimage import binary_dilation
from skimage.morphology import binary_erosion, ball
from scipy.ndimage import label as label_ndimage
import statistics
import pandas as pd
from collections import Counter

import jwt
import xnat
from barbell2_bodycomp.convert import dcm2nifti
from pydicom import dcmread
from fastapi import FastAPI, Body, Depends, HTTPException, status, File, UploadFile
from fastapi.responses import FileResponse, Response
from fastapi.security import OAuth2PasswordBearer, OAuth2PasswordRequestForm
from passlib.hash import bcrypt
from tortoise import fields
from tortoise.models import Model
from tortoise.contrib.fastapi import register_tortoise
from tortoise.contrib.pydantic import pydantic_model_creator
from pydantic import BaseModel, Field
from typing import Annotated
import nibabel as nib

logging.basicConfig(format='%(asctime)s :: %(levelname)s :: %(funcName)s :: %(lineno)d :: %(message)s', level = logging.INFO)

# openssl rand -hex 32
JWT_SECRET = '5c31c340b35071a2970f7b9b717c7e3e'
JWT_ALGORITHM = 'HS256'
app = FastAPI()

def update_cache():
    while True:
        if not(os.path.exists(".cache")):
            print("Creating .cache file")
            with open(".cache", "w") as f:
                secret = secrets.token_hex(8)
                f.write(secret)
                print("Secret: {} 1".format(secret))
                f.write(" ")
                f.write("1")
            continue
        with open(".cache", "r+") as f:
            data = f.read().split(" ")
        if time.time() - float(data[1]) > 60*60*24*10:             # 10 days
            with open(".cache", "w") as f:
                secret = secrets.token_hex(8)
                f.write(secret)
                f.write(" ")
                time_now = time.time()
                f.write(str(time_now))
                print("Secret: {} {}".format(secret, time_now))
        time.sleep(60*60*24)

update_cache_thread = threading.Thread(name="update_cache", target=update_cache)
update_cache_thread.start()

class User(Model):
    id = fields.IntField(pk=True)
    username = fields.CharField(50, unique=True)
    password_hash = fields.CharField(128)
    permission_key = fields.CharField(128)

    @classmethod
    async def get_user(cls, username):
        return cls.get(username=username)

    def verify_password(self, password):
        return bcrypt.verify(password, self.password_hash)

class Vertebrae(str, Enum):
    S1 = "S1"
    L5 = "L5"
    L4 = "L4"
    L3 = "L3"
    L2 = "L2"
    L1 = "L1"
    T12 = "T12"
    T11 = "T11"
    T10 = "T10"
    T9 = "T9"
    T8 = "T8"
    T7 = "T7"
    T6 = "T6"
    T5 = "T5"
    T4 = "T4"
    T3 = "T3"
    T2 = "T2"
    T1 = "T1"
    C7 = "C7"
    C6 = "C6"
    C5 = "C5"
    C4 = "C4"
    C3 = "C3"
    C2 = "C2"
    C1 = "C1"

class RequestBody(BaseModel):
    project_name: str
    xnat_url: str
    xnat_username: str
    xnat_password: str
    vertebra: Vertebrae

User_Pydantic = pydantic_model_creator(User, name="User")
# RequestBody_Pydantic = pydantic_model_creator(RequestBody, name="RequestBody")
UserIn_Pydantic = pydantic_model_creator(User, name="UserIn", exclude_readonly=True)

oauth2_scheme = OAuth2PasswordBearer(tokenUrl="token")

class MicroManager:
    def __init__(self):
        self.UPLOAD_DIR = "temp"
        if not(os.path.exists(self.UPLOAD_DIR)):
            os.mkdir(self.UPLOAD_DIR)
        self.SLICES_DIR = os.path.join(self.UPLOAD_DIR, "slices")
        self.NIFTI_DIR = os.path.join(self.UPLOAD_DIR, "nifti")
        if not(os.path.exists(self.NIFTI_DIR)):
            os.mkdir(self.NIFTI_DIR)
        self.TS_DIR = os.path.join(self.UPLOAD_DIR, "TS_outputs")
        if not(os.path.exists(self.TS_DIR)):
            os.mkdir(self.TS_DIR)
        self.PROJECT_DIR = "/Users/abhimanyu/Maastricht/Internship/XNAT-TS-Mosamatic-Microservice"

        self.session = None
        self.xnat_project_name = None
        self.thread_complete = True
        self.error_subjects = {}
        self.data_downloaded = True
        self.converted_first_dicom = False

    def cleanup(self):
        shutil.rmtree(self.UPLOAD_DIR)
        os.mkdir(self.UPLOAD_DIR)

    def TotalSegmentator(self, input_file: str, output_path: str, subject: str):
        try:
            if not os.path.exists(output_path):
                os.makedirs(output_path, exist_ok=True)
            cmd = f'TotalSegmentator -i {input_file} -o {output_path} --fast'
            logging.info("TS cmd: {}".format(cmd))
            os.system(cmd)
            output_path = os.path.join(output_path, subject)
            cmd = f'TotalSegmentator -i {input_file} -o {output_path} --ml --fast'
            logging.info("TS cmd: {}".format(cmd))
            os.system(cmd)
            return 200, "Success"
        except Exception as e:
            logging.error("Error: {}".format(e))
            return 501, f"Error in microManager::TotalSegmentator: {e}"

    def sliceSelector(self, subject, TS_path, vertebra: Vertebrae = Vertebrae.L3):
        try:
            FILENAME = "vertebrae_" + vertebra + ".nii.gz"
            PATH = os.path.join(TS_path, FILENAME)
            if not(os.path.exists(self.SLICES_DIR)):
                os.mkdir(self.SLICES_DIR)
            filename = f"{subject}_vertebrae_{vertebra.value}.nii.gz"
            shutil.copyfile(PATH, os.path.join(self.SLICES_DIR, filename))
            return 200, "Success"
        except Exception as e:
            logging.error("Error: {}".format(e))
            return 502, f"Error in sliceSelector: {e}"

    def deleteTheRest(self, TS_path):
        try:
            shutil.rmtree(TS_path)
            return 200, "Success"
        except Exception as e:
            logging.error("Error: {}".format(e))
            return 503, f"Error in microManager::deleteTheRest: {e}"

    def check_segmentation(self, SEG_PATH):
        nii_file = nib.load(SEG_PATH)
        seg = nii_file.get_fdata().astype(int)
        self.check_orientation(seg)
        seg = self.check_vertebrae_masks_consistency(seg)
        self.save_vertebrae_bounding_boxes(seg)
        return 200, "Success"

    def check_vertebrae_masks_consistency(self, seg, radius=2.5):
        try:
            logging.info("----------------------------------------------------------------")
            logging.info(" ---------------------- MASK CHECK -----------------------------")
            logging.info("----------------------------------------------------------------")

            classes_dic = class_map['total']
            logging.info("classes_dic: {}".format(classes_dic))

            # Apply morphological operations to the mask
            # This is done to seperate the segmentation masks of each vertebrae
            struct_elem = ball(radius)  # choose appropriate size and shape of structuring element

            new_seg = np.zeros_like(seg)

            spine = {}

            logging.info("Applying morphological transformations...")

            for label_value in range(18, 42):
                # Create a binary array for the current label value
                # Morphological operations are applied one label at a time
                binary_array = (seg == label_value)
                if np.sum(binary_array) > 0:
                    spine[label_value] = np.sum(binary_array)
                else:
                    continue
                # Apply erosion to the segmentation array
                eroded_array = binary_erosion(binary_array, struct_elem)
                # Apply dilation to the eroded segmentation array
                dilated_array = binary_dilation(eroded_array, struct_elem)
                # Create a binary mask of the dilated segmentation array
                new_seg[dilated_array] = label_value

            logging.info("Separating each 3D mask...")

            # We only keep the labels of the vertebrae
            mask = np.logical_or(seg < 18, seg > 41)
            new_seg[mask] = seg[mask]
            s = new_seg
            # print("mask", mask.shape)
            # print("seg", seg.shape)
            # print("seg[mask]", seg[mask].shape)
            # print("new_seg[mask]", new_seg[mask].shape)
            # with open('seg.npy', 'wb') as f:
            #     np.save(f, seg)
            # with open('new_seg.npy', 'wb') as f:
            #     np.save(f, new_seg)

            # We compute the average size of a vertebra mask
            # This is to distinguish between masks of full vertebrae or "isolated islands"
            # We trim some values to not include the extremes in the average
            # calculate the number of values to trim from both ends

            trim_percentage = 0.1 # define the percentage of values to trim
            trim_count = round(len(spine) * trim_percentage)
            # print("len(spine)", len(spine))
            if trim_count < 1: # trim at least the min and max val
                trim_count = 1
            # print("trim_count", trim_count)
            # sort the values
            sorted_values = sorted(spine.values())
            # print("sorted_values", sorted_values)
            # trim the values
            trimmed_values = sorted_values[trim_count:len(sorted_values) - trim_count]
            # calculate the mean of the trimmed values
            if trimmed_values:
                trimmed_mean = statistics.mean(trimmed_values)
            else:
                trimmed_mean = 0
            # print("trimmed_mean", trimmed_mean)

            mask_arr = np.logical_and(s >= 18, s <= 41)

            # Perform connected component analysis on the mask
            labels, num_labels = label_ndimage(mask_arr)
            # print("num_labels", num_labels)
            # print("labels", labels.shape)

            label_assigned = {}

            logging.info('Checking for inconsistencies in labeling ...')

            # Check each identified 3D mask
            for i in range(1, num_labels + 1):
                mask = (labels == i)
                indices = np.argwhere(mask)
                center = np.mean(indices, axis=0)
                vertebra_labels = s[mask]
                # Count each unique label in the 3D label
                label_counts = Counter(vertebra_labels)
                if len(label_counts) > 1:
                    # Mask contains pixels from multiple vertebrae labels, keep only the largest connected component for the correct label
                    correct_label = max(label_counts, key=label_counts.get)
                    logging.info("Multiple labels are assigned to the same segmentation mask as {}".format(str(classes_dic[correct_label])))
                else:
                    correct_label = list(label_counts.keys())[0]
                # if the label has been previously assigned, we detect what the most plausible error is
                if correct_label in label_assigned:
                    if label_counts[correct_label] < trimmed_mean / 4:
                        logging.info("{} appear more than once. A small seperate section is given this label.".format(str(classes_dic[correct_label])))
                    else:
                        if center[0] < label_assigned[correct_label]:
                            if correct_label < 18:
                                logging.info("{} appear more than once. It might have needed to be assigned the saccrum label.".format(str(classes_dic[correct_label])))
                            else:
                                logging.info("{} appear more than once. It might have needed to be assigned the previous vertebra label.".format(str(classes_dic[correct_label])))
                        else:
                            logging.info("{} appear more than once. It might have needed to be assigned the following vertebra label.".format(str(classes_dic[correct_label])))
                else:
                    # Add the current label to the list of assigned labels
                    label_assigned[correct_label] = center[0]
                logging.info("Check Done")
            return new_seg
        except Exception as e:
            logging.error("Error: {}".format(e))
            return seg

    # Function to save the coordinates of each vertebrae labeled by Total segmentator and ensure the results are consistent
    # Input: seg - the label array of TotalSegmentator
    # Output: None (saves a csv)
    def save_vertebrae_bounding_boxes(self, seg):
        # internal function to define the borders of the bounding box
        def calculate_bounding_box(label):
            indices = np.argwhere(label != 0)
            if len(indices) == 0:
                return None

            min_coords = min(indices, key=lambda sublist: sublist[2])
            max_coords =  max(indices, key=lambda sublist: sublist[2])
            center_coords = (min_coords + max_coords) // 2
            volume = np.count_nonzero(label)

            return {
                "min_coords [z x y]": min_coords,
                "max_coords [z x y]": max_coords,
                "center_coords [z x y]": center_coords,
                "volume": volume
            }
        # iterate through the vertebrae to extract coordinates and check correct order
        labeled_vertebrae = []
        for label in range(18, 42):
            vertebra_label = (seg == label)
            if np.any(vertebra_label):
                bounding_box_info = calculate_bounding_box(vertebra_label)
                if bounding_box_info is not None:
                    if len(labeled_vertebrae) > 0:
                        if label != labeled_vertebrae[-1]["label"] + 1:
                            logging.info("A vertebra label is missing: {}".format(str(labeled_vertebrae[-1]["label"] + 1)))
                            # print("label: " + str(label))
                            # print("labeled_vertebrae[-1][\"label\"] + 1: " + str(labeled_vertebrae[-1]["label"] + 1))
                    labeled_vertebrae.append({
                        "label": label,
                        **bounding_box_info
                    })

        logging.info("{} vertebrae are labeled on this CT image".format(str(len(labeled_vertebrae))))

        df = pd.DataFrame(labeled_vertebrae)
        df.to_csv("vertebrae_info.csv", index=False)

        for index, row in df.iterrows():
            if index > 0:
                if row["center_coords [z x y]"][2] < prev:
                    logging.info("{} doesnt appear in correct position".format(str(row["label"])))
            prev = row["center_coords [z x y]"][2]

    # Function that checks the output of Total segmentator to ensure it wasn't given an image upside down
    # Input: seg - the label array of TotalSegmentator, precision - how many slice tolerance to check order of labels
    # Output: None (prints results on terminal)
    def check_orientation(self, seg, precision=5):
        logging.info("----------------------------------------------------------------")
        logging.info(" ---------------- ORIENTATION CHECK ----------------------------")
        logging.info("----------------------------------------------------------------")

        wrong_orientation = False

        # Each organ and bones is assigned a level to be able to check wheter the segmentation appear in the correct order
        map_organs_to_level = {'brain': 1, 'heart_myocardium': 2, 'heart_atrium_left': 2, 'heart_atrium_right': 2,
                            'heart_ventricle_left': 2
            , 'heart_ventricle_right': 2, 'stomach': 3, 'urinary_bladder': 4}
        map_bones_to_lvl = {'vertebrae_C2': 2, 'vertebrae_C4': 4, 'vertebrae_C6': 6, 'vertebrae_T1': 8, 'vertebrae_T3': 10
            , 'vertebrae_T5': 12, 'vertebrae_T7': 14, 'vertebrae_T9': 16, 'vertebrae_T11': 18, 'vertebrae_L1': 20,
                            'vertebrae_L3': 22
            , 'vertebrae_L5': 24, 'hip_right': 24, 'hip_left': 24, 'sacrum': 24}

        logging.info("Iterating through slices ...")

        classes_dic = class_map["total"]
        # First we loop through the scan to check the organs are correct then we check the bones
        for i in range(2):
            if i == 0:
                map_classes_to_lvl = map_organs_to_level
            if i == 1:
                map_classes_to_lvl = map_bones_to_lvl
            start_index = len(seg[0][0])
            # print("start_index", start_index)
            # print("seg", seg.shape)
            map_lvl_to_classes = {}
            for k, v in map_classes_to_lvl.items():
                map_lvl_to_classes[v] = map_lvl_to_classes.get(v, []) + [k]
            # We loop every 5 slice, for each slice we keep in memory the segmentation that is supposed to be at the "highest level"
            length = seg.shape[2]
            for y in range(length):
                if start_index - precision > precision:
                    start_index = start_index - precision
                    # print("start_index", start_index)
                    slice = seg[:, :, start_index]
                    # slice = seg[start_index]
                    # from the slice we store every unique label
                    count = Counter(slice.flatten())
                    slice = set(slice.flatten())
                    # we remove no label ie 0
                    slice.remove(0)
                    arr = []
                    for el in slice:
                        if count[el] > 20:
                            cat = classes_dic[el]
                            arr.append(cat)
                    # we only keep labels that have been predefined a level to check if they appear in the correct order
                    arr = [i for i in arr if i in map_classes_to_lvl]
                    if len(arr) > 0:
                        # we find out of each label which is at max assigned level
                        max_lvl1 = 1
                        for cat in arr:
                            if map_classes_to_lvl[cat] > max_lvl1:
                                max_lvl1 = map_classes_to_lvl[cat]
                        index = start_index
                        # We loop every x (precision) slice appearing after the slice being currently checked. To make sure every segmentation is at a lower or equal level as the max level identified in slice 1
                        for x in range(start_index - precision):
                            # we repeat the same process on the following slices and compare to max_lvl1
                            if index - precision > 0:
                                index = index - precision
                                slice = seg[:, :, index]
                                # slice = seg[index]
                                count = Counter(slice.flatten())
                                slice = set(slice.flatten())
                                slice.remove(0)
                                arr = []
                                for el in slice:
                                    if count[el] > 20:
                                        cat = classes_dic[el]
                                        arr.append(cat)
                                arr = [i for i in arr if i in map_classes_to_lvl]
                                if len(arr) > 0:
                                    max_lvl2 = 1
                                    for cat in arr:
                                        if map_classes_to_lvl[cat] > max_lvl2:
                                            max_lvl2 = map_classes_to_lvl[cat]
                                    # If 1 of the segmentation is at a lower level, we print a warning of this
                                    if max_lvl2 < max_lvl1:
                                        wrong_orientation = True
                                        logging.info("{}".format(map_lvl_to_classes[max_lvl2]))
                                        logging.info('appear after')
                                        logging.info("{}".format(map_lvl_to_classes[max_lvl1]))
                                        logging.info('----')
                                        break
                        if wrong_orientation:
                            break
        if wrong_orientation:
            logging.info("Input CT scan may not be in the correct orientation")

    def manager(self, vertebra):
        try:
            self.data_downloaded, project_dir = self.getFromXNAT()
            logging.info("project_dir: {}".format(project_dir))
            if not self.data_downloaded:
                return

            # dicomToNifti_thread = threading.Thread(name="dicomToNifti", target=self.dicomToNifti, args=(project_dir,))
            # dicomToNifti_thread.start()

            #TODO: Remove this while loop and replace with thread lock variable thingy
            # while True:
            #     if self.converted_first_dicom:
            #         break
            #     time.sleep(2)

            self.thread_complete = False
            for subject in os.listdir(project_dir):
                status = 200
                msg = ""
                try:
                    self.dicomToNifti(os.path.join(project_dir, subject))
                    input_file = os.path.join(self.NIFTI_DIR, subject, f"{subject}.nii.gz")
                    TS_path = os.path.join(self.TS_DIR, f"{subject}")

                    logging.info("input_file: {}".format(input_file))
                    logging.info("TS_path: {}".format(TS_path))

                    status, msg = self.TotalSegmentator(input_file, TS_path, subject)
                    if status == 501:
                        self.error_subjects[subject] = {}
                        self.error_subjects[subject]["status"] = status
                        self.error_subjects[subject]["msg"] = msg
                        continue
                    logging.info("TS done")

                    logging.info("starting seg check")
                    seg_path = os.path.join(TS_path, f"{subject}.nii")
                    logging.info("seg_path: {}".format(seg_path))
                    status, msg = self.check_segmentation(seg_path)
                    if status == 501:
                        self.error_subjects[subject] = {}
                        self.error_subjects[subject]["status"] = status
                        self.error_subjects[subject]["msg"] = msg
                        continue
                    logging.info("checking segmentation done")

                    status, msg = self.sliceSelector(subject, TS_path, vertebra)
                    if status == 501:
                        self.error_subjects[subject] = {}
                        self.error_subjects[subject]["status"] = status
                        self.error_subjects[subject]["msg"] = msg
                    logging.info("slice selection done")

                    status, msg = self.deleteTheRest(TS_path)
                    if status == 501:
                        self.error_subjects[subject] = {}
                        self.error_subjects[subject]["status"] = status
                        self.error_subjects[subject]["msg"] = msg
                        continue
                    logging.info("deleting the rest done")

                except Exception as e:
                    self.error_subjects[subject] = {}
                    self.error_subjects[subject]["status"] = 501
                    self.error_subjects[subject]["msg"] = f"Error in microManager::manager: {e}"
                    continue
            self.thread_complete = True
        except Exception as e:
            logging.error("Error: {}".format(e))
        return

    def checkDataReadiness(self, background_tasks: BackgroundTasks):
        if not self.data_downloaded:
            return Response(
                status_code=status.HTTP_424_FAILED_DEPENDENCY,
                content="Download from XNAT was unsuccessful. Please try again",
                media_type="application/xml"
            )
        if os.path.exists(self.SLICES_DIR):
            if not self.thread_complete:
                return Response(
                    status_code=status.HTTP_202_ACCEPTED,
                    content="Data is not ready yet.",
                    media_type="application/xml"
                )
            else:
                shutil.make_archive(self.SLICES_DIR, 'zip', self.SLICES_DIR)
                zip_path = os.path.join(self.UPLOAD_DIR, "slices.zip")
                background_tasks.add_task(self.cleanup)
                return FileResponse(
                    path=zip_path,
                    media_type="application/octet-stream",
                    filename="slices.zip",
                    status_code=status.HTTP_200_OK
                )
        else:
            return Response(
                status_code=status.HTTP_404_NOT_FOUND,
                content="First process the data, then ask for the slices.",
                media_type="application/xml"
            )

    def checkXNAT(self, project_name: str, xnat_url: str, xnat_username: str, xnat_password: str):
        try:
            try:
                self.session = xnat.connect(xnat_url, user=xnat_username, password=xnat_password, default_timeout=30)
            except:
                return Response(
                    status_code=status.HTTP_401_UNAUTHORIZED,
                    content="Invalid credentials. Check the XNAT URL, username and password (Note: You need to be connected to the UM VPN)",
                    media_type="application/xml"
                )
            # session = xnat.connect('http://137.120.191.233', user='abhi', password='welcome2023', default_timeout=30)

            if not project_name in self.session.projects:
                return Response(
                    status_code=status.HTTP_404_NOT_FOUND,
                    content="The project you were looking for was not found in the XNAT library",
                    media_type="application/xml"
                )

            self.xnat_project_name = project_name

            return Response(
                status_code=status.HTTP_200_OK,
                content="XNAT Data validated",
                media_type="application/xml"
            )
        except Exception as e:
            logging.error("Error:{}".format(e))
            return Response(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                content=f"Error in microManager::checkXNAT: {e}",
                media_type="application/xml"
            )

    def getFromXNAT(self):
        try:
            project = self.session.projects[self.xnat_project_name]
            project.download_dir(self.UPLOAD_DIR)
            return True, os.path.join(self.UPLOAD_DIR, self.session.projects[self.xnat_project_name].name)
        except:
            return False, ""

    def getDicomFolder(self, dir):
        rootdir = dir
        for subdir, dirs, files in os.walk(rootdir):
            for file in files:
                if file.endswith('.dcm'):
                    dcm_filename = os.path.join(subdir, file).split('/')[-1]
                    dcm_folder = os.path.join(subdir, file).replace(dcm_filename, "")
                    return dcm_folder

    def dicomToNifti(self, project_dir: str):
        try:
            # for folder in os.listdir(project_dir):
            #     dcm_folder = self.getDicomFolder(os.path.join(project_dir, folder))
            #     dcm_obj = dcm2nifti.DicomToNifti()
            #     dcm_obj.input_directory = dcm_folder
            #     dcm_obj.output_file = os.path.join(self.NIFTI_DIR, folder, f"{folder}.nii.gz")
            #     dcm_obj.execute()
            #     self.converted_first_dicom = True

            dcm_folder = self.getDicomFolder(project_dir)
            logging.info("project_dir: {}".format(project_dir))
            dcm_obj = dcm2nifti.DicomToNifti()
            dcm_obj.input_directory = dcm_folder
            logging.info("dcm_folder: {}".format(dcm_folder))
            subject = project_dir.split('/')[-1]
            dcm_obj.output_file = os.path.join(self.NIFTI_DIR, subject, f"{subject}.nii.gz")
            logging.info("output_file: {}".format(dcm_obj.output_file))
            dcm_obj.execute()
            self.converted_first_dicom = True
            return True
        except Exception as e:
            logging.error("Error:".format(e))
            return False

async def authenticate_user(username: str, password: str):
    user = await User.get(username=username)
    if not user:
        return False
    if not user.verify_password(password):
        return False
    return user

async def get_current_user(token: str = Depends(oauth2_scheme)):
    try:
        payload = jwt.decode(token, JWT_SECRET, algorithms=[JWT_ALGORITHM])
        user = await User.get(id=payload.get('id'))
    except:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Invalid username or password"
        )
    return await User_Pydantic.from_tortoise_orm(user)

@app.post('/users', response_model=User_Pydantic)
async def create_user(user: UserIn_Pydantic):
    # open and read .cache file and get the 16 digit key and compare it with the key provided by the user
    with open(".cache", "r") as f:
        permission_key = f.read()
        # the permission key is the first 16 characters of the .cache file
        permission_key = permission_key[:16]
    print("Permission key: {}".format(permission_key))
    print("User permission key: {}".format(user.permission_key))
    if permission_key != user.permission_key:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Invalid permission key"
        )
    user_obj = User(username=user.username, password_hash=bcrypt.hash(user.password_hash), permission_key=user.permission_key)
    await user_obj.save()
    print("User saved")
    user_serialized = await User_Pydantic.from_tortoise_orm(user_obj)
    print("User serialized")
    return user_serialized

@app.get('/user/me', response_model=User_Pydantic)
async def get_user(user: User_Pydantic = Depends(get_current_user)):
    username = user.username
    id = user.id
    return Response(
        status_code=status.HTTP_200_OK,
        content=f"Username: {username}, ID: {id}",
        media_type="application/xml"
    )

@app.post('/token')
async def generate_token(form_data: OAuth2PasswordRequestForm = Depends()):
    user = await authenticate_user(form_data.username, form_data.password)
    if not user:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Invalid username or password"
        )
    user_obj = await User_Pydantic.from_tortoise_orm(user)
    token = jwt.encode(user_obj.dict(), JWT_SECRET, algorithm=JWT_ALGORITHM)
    return {"access_token": token, "token_type": "bearer"}

# Create an API to get an input dicom file from the user and return the output dicom file
# @router.post("/dicom")
# async def dicom(input_dicom: UploadFile = File, current_user: User_Pydantic = Depends(get_current_user)):
#     ds = dcmread(input_dicom.file, force=True)
#     SAVE_FILE_PATH = os.path.join(UPLOAD_DIR, input_dicom.filename)
#     ds.save_as(SAVE_FILE_PATH)
#     return FileResponse(path=SAVE_FILE_PATH, media_type="application/octet-stream", filename=input_dicom.filename)

@app.post("/ProjectName")
async def from_mosamatic (
        background_tasks: BackgroundTasks,
        request: Annotated[RequestBody, Body(embed=True)],
        current_user: User_Pydantic = Depends(get_current_user)
    ):
    global microObject
    microObject = MicroManager()
    response = microObject.checkXNAT(request.project_name, request.xnat_url, request.xnat_username, request.xnat_password)
    if response.status_code != 200:
        return response

    microManager_thread = threading.Thread(name="microManager", target=microObject.manager, args=(request.vertebra,))
    microManager_thread.start()

    return Response(
        status_code=status.HTTP_202_ACCEPTED,
        content="Received the request. Please wait for the results.",
        media_type="application/xml"
    )

@app.post("/isDataReady")
async def is_data_ready (background_tasks: BackgroundTasks, current_user: User_Pydantic = Depends(get_current_user)):
    global microObject
    try:
        response = microObject.checkDataReadiness(background_tasks)
    except Exception as e:
        response = Response(
            status_code=status.HTTP_404_NOT_FOUND,
            content="Process the data first",
            media_type="application/xml"
        )
    return response

register_tortoise(
    app,
    db_url="sqlite://db.sqlite3",
    modules={"models": ["main"]},
    generate_schemas=True,
    add_exception_handlers=True
)



from PySide6.QtWidgets import QGraphicsItemGroup
import pydicom
import pydicom.errors
import numpy as np

from PySide6.QtGui import QImage, QPixmap
from PySide6.QtWidgets import QGraphicsItemGroup, QGraphicsPixmapItem

from mosamaticdesktop.widgets.viewers.dicomviewer.layer import Layer
from mosamaticdesktop.utils import readFromCache, writeToCache, applyWindowCenterAndWidth


class DicomLayer(Layer):
    def __init__(self) -> None:
        super(DicomLayer, self).__init__(name='dicom', opacity=1.0)
        self._qimage = None
        self._windowLevel = 50
        self._windowWidth = 400

    def data(self) -> pydicom.FileDataset:
        if self.file():
            content = readFromCache(file=self.file())
            if not content:
                data = pydicom.dcmread(self.file().path())
                data.decompress()
                content = writeToCache(file=self.file(), fileObject=data)
            data = content.fileObject()
            return data
        return None
        
    def setWindowLevelAndWidth(self, windowLevel: int, windowWidth: int) -> None:
        self._windowLevel = windowLevel
        self._windowWidth = windowWidth

    def convertDataToQImage(self) -> QImage:
        p = self.data()
        pixelArray = p.pixel_array
        if 'RescaleSlope' in p and 'RescaleIntercept' in p:
            pixelArray = pixelArray * p.RescaleSlope + p.RescaleIntercept
        pixelArray = applyWindowCenterAndWidth(pixelArray, self._windowLevel, self._windowWidth)
        if pixelArray.dtype != np.uint8:
            pixelArray = pixelArray.astype(np.uint8)
        height, width = pixelArray.shape
        bytes_per_line = width
        return QImage(pixelArray.data, width, height, bytes_per_line, QImage.Format_Grayscale8)

    def createGraphicsItem(self) -> QGraphicsItemGroup:
        group = QGraphicsItemGroup()
        if self.file() and not self._qimage:
            self._qimage = self.convertDataToQImage()
        if self._qimage and self.visible():
            pixmap = QPixmap.fromImage(self._qimage)
            pixmapItem = QGraphicsPixmapItem(pixmap)
            pixmapItem.setOpacity(self.opacity())
            group.addToGroup(pixmapItem)
        return group


import os
import sys
import logging

from typing import List

from mosamaticdesktop.singleton import singleton
from mosamaticdesktop.utils import createNameWithTimestamp

LOGFILEPATH = os.environ.get('LOGFILEPATH', 'MosamaticDesktop.log')


@singleton
class Logger:
    def __init__(self) -> None:
        # if os.path.isfile(LOGFILEPATH):
        #     os.remove(LOGFILEPATH)
        # with open(LOGFILEPATH, 'w') as f:
        #     pass
        self._logger = logging.getLogger('MosamaticDesktop')
        self._logger.addHandler(self.standardOutputHandler())
        # self._logger.addHandler(self.fileOutputHandler())
        # This prevent logging to be propagated up to the root logger
        # which TensorFlow might reconfigure
        self._logger.propagate = False 
        self.enableInfo()

    def standardOutputHandler(self) -> logging.StreamHandler:
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(formatter)
        return handler
    
    def fileOutputHandler(self) -> logging.FileHandler:
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        handler = logging.FileHandler(LOGFILEPATH)
        handler.setFormatter(formatter)
        return handler
    
    def logFilePath(self) -> str:
        print(f'Logger.logFilePath() {LOGFILEPATH}')
        return LOGFILEPATH
    
    def logHandlers(self) -> List[logging.Handler]:
        return self._logger.handlers
    
    def enableDebug(self) -> None:
        self._logger.setLevel(logging.DEBUG)
        for handler in self._logger.handlers:
            handler.setLevel(logging.DEBUG)

    def enableInfo(self) -> None:
        self._logger.setLevel(logging.INFO)
        for handler in self._logger.handlers:
            handler.setLevel(logging.INFO)
    
    def info(self, message: str) -> None:
        self._logger.info(message)

    def warning(self, message: str) -> None:
        self._logger.warning(message)

    def error(self, message: str) -> None:
        self._logger.error(message)

    def debug(self, message: str) -> None:
        self._logger.debug(message)


from mosamaticdesktop.tasks.taskwidget import TaskWidget
from mosamaticdesktop.tasks.createpngfromdicomfiletask.createpngfromdicomfiletask import CreatePngFromDicomFileTask


class CreatePngFromDicomFileTaskWidget(TaskWidget):
    def __init__(self) -> None:
        super(CreatePngFromDicomFileTaskWidget, self).__init__(taskType=CreatePngFromDicomFileTask)
    
    def validate(self) -> None:
        pass


import numpy as np

from typing import List

from PySide6.QtGui import QImage, QPixmap
from PySide6.QtWidgets import QGraphicsItemGroup, QGraphicsPixmapItem

from mosamaticdesktop.widgets.viewers.dicomviewer.layer import Layer
from mosamaticdesktop.utils import convertNumPyArrayToRgbQImage, AlbertaColorMap, readFromCache, writeToCache, tagPixels


class TagLayer(Layer):
    def __init__(self) -> None:
        super(TagLayer, self).__init__(name='tag')
        self._qimage = None
        self._shape = None

    def shape(self) -> List[int]:
        return self._shape

    def setShape(self, shape: List[int]) -> None:
        self._shape = shape

    def data(self) -> np.array:
        if self.file():
            content = readFromCache(file=self.file())
            if not content:
                data = tagPixels(tagFilePath=self.file().path())
                content = writeToCache(file=self.file(), fileObject=data)
            data = content.fileObject()
            data = data.reshape(self.shape())
            return data
        return None

    def convertDataToQImage(self) -> QImage:
        return convertNumPyArrayToRgbQImage(numpyArray=self.data(), colorMap=AlbertaColorMap())

    def createGraphicsItem(self) -> QGraphicsItemGroup:
        group = QGraphicsItemGroup()
        if self.file() and not self._qimage:
            self._qimage = self.convertDataToQImage()
        if self._qimage:
            pixmap = QPixmap.fromImage(self._qimage)
            pixmapItem = QGraphicsPixmapItem(pixmap)
            pixmapItem.setOpacity(self.opacity())
            group.addToGroup(pixmapItem)
        return group


from typing import List

from PySide6.QtCore import QPoint
from PySide6.QtWidgets import QMenu, QTreeView, QMessageBox

from mosamaticdesktop.widgets.filesetitem import FileSetItem
from mosamaticdesktop.data.datamanager import DataManager


class MultiFileSetItemMenu(QMenu):
    def __init__(self, treeView: QTreeView, fileSetItems: List[FileSetItem], globalPos: QPoint) -> None:
        super(MultiFileSetItemMenu, self).__init__()
        self._treeView = treeView
        self._fileSetItems = fileSetItems
        self._position = globalPos
        self._dataManager = DataManager()
        deleteFileSetAction = self.addAction('Delete')
        deleteFileSetAction.triggered.connect(self.deleteFileSet)

    def deleteFileSet(self):
        for fileSetItem in self._fileSetItems:
            if isinstance(fileSetItem, FileSetItem):
                fileSet = fileSetItem.fileSet()
                self._dataManager.deleteFileSet(fileSet=fileSet)

    def show(self):
        self.exec_(self._position)


from sqlalchemy.orm import sessionmaker

from mosamaticdesktop.data.engine import Engine


class Session:
    def __init__(self) -> None:
        session = sessionmaker(bind=Engine().get())
        self._session = session()

    def __enter__(self):
        return self._session
    
    def __exit__(self, exc_type, exc_value, traceback):
        self._session.close()



import requests
import time
import sys
import argparse
from getpass import getpass

def main():
    arg = argparse.ArgumentParser()
    arg.add_argument("--new-user", default=False, action=argparse.BooleanOptionalAction, required=False, help="Create a new user")
    arg.add_argument("--existing-user", default=True, action=argparse.BooleanOptionalAction, required=False, help="Use an existing user")
    args = arg.parse_args()

    print("new-user: ", args.new_user)
    print("existing-user: ", args.existing_user)

    if args.new_user:
        create_user_url = 'http://127.0.0.1:8000/users'
        premission_key = input("Enter permission key: ")
        username = input("Create new username: ")
        password = getpass("Create password: ")
        re_password = getpass("Re-enter password: ")
        if password != re_password:
            print("Passwords do not match")
            exit(1)
        data = {
                "username": username,
                "password_hash": password,
                "permission_key": premission_key
            }

        header = {
            "accept": "application/json",
            "Content-Type": "application/json"
        }
        response = requests.post(create_user_url, headers=header, json=data)
        print(f"Response: {response.status_code}, {response.content.decode('utf-8')}")
    elif args.existing_user:
        username = input("Enter username: ")
        password = getpass("Enter password: ")

    auth_url = "http://127.0.0.1:8000/token"
    data = {
        "grant_type": "",
        "username": username,
        "password": password,
        "scope": "",
        "client_id": "",
        "client_secret": ""
    }

    # Send the first request to obtain the JWT token
    response = requests.post(auth_url, data=data, headers={"accept": "application/json", "Content-Type": "application/x-www-form-urlencoded"})

    if response.status_code == 200:
        # Extract the JWT token from the response
        response_data = response.json()
        jwt_token = response_data.get("access_token")
    else:
        print(f"Error: {response.status_code}, {response.content.decode('utf-8')}")

    url_params = {
        "request": {
            "project_name": "TD",
            "xnat_url": "http://137.120.191.233",
            "xnat_username": "abhi",
            "xnat_password": "welcome2023",
            "vertebra": "L3"
        }
    }

    # Send the second request to start the data processing
    request_url = "http://127.0.0.1:8000/ProjectName"
    header = {
        "accept": "application/json",
        "Authorization": f"Bearer {jwt_token}",
        "Content-Type": "application/json"
    }
    response = requests.post(request_url, headers=header, json=url_params)
    print(f"Response: {response.status_code}, {response.content.decode('utf-8')}")
    if not response.status_code == 202:
        exit(1)

    # Replace this URL with the actual URL of your FastAPI server
    get_url = "http://localhost:8000/isDataReady"
    # JSON payload for the POST request
    headers = {
        "Authorization": f"Bearer {jwt_token}"
    }

    while True:
        try:
            response = requests.post(get_url, headers=headers)
            if response.status_code == 200:
                # If you expect to receive a file response, you can save it to a file
                with open("slices.zip", "wb") as file:
                    file.write(response.content)
                print("Data is ready and saved as slices.zip")
                break
            else:
                print(f"{response.status_code} {response.content.decode('utf-8')}")
        except requests.exceptions.RequestException as e:
            print(f"Request Exception: {e}")
        time.sleep(30)

if __name__ == "__main__":
    main()


from mosamaticdesktop.tasks.taskwidget import TaskWidget
from mosamaticdesktop.tasks.createpngfromtagfiletask.createpngfromtagfiletask import CreatePngFromTagFileTask


class CreatePngFromTagFileTaskWidget(TaskWidget):
    def __init__(self) -> None:
        super(CreatePngFromTagFileTaskWidget, self).__init__(taskType=CreatePngFromTagFileTask)
    
    def validate(self) -> None:
        pass


#!/usr/bin/env python

import os

from setuptools import setup, find_packages

requirements = []
with open('requirements.txt', 'r') as f:
    for line in f.readlines():
        line = line.strip()
        requirements.append(line)

setup_requirements = []

test_requirements = []

setup(
    author="Ralph Brecheisen",
    author_email='r.brecheisen@maastrichtuniversity.nl',
    python_requires='>=3.11',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    # package_dir={
    #     'src': 'src/mosamaticdesktop',
    # },
    # package='mosamaticdesktop',
    packages=find_packages(include=[
        'mosamaticdesktop', 
        'mosamaticdesktop.*', 
        'mosamaticdesktop.aiservice',
        'mosamaticdesktop.data',
        'mosamaticdesktop.data.models',
        'mosamaticdesktop.tasks',
        'mosamaticdesktop.tasks.*',
        'mosamaticdesktop.widgets',
        'mosamaticdesktop.widgets.*',
    ]),
    include_package_data=True,
    package_data={
        'mosamaticdesktop': ['scripts/*', 'VERSION', 'GIT_COMMIT_ID'],
    },
    description="Desktop tool for analyzing medical images",
    install_requires=requirements,
    license="MIT license",
    keywords='mosamaticdesktop',
    name='mosamaticdesktop',
    setup_requires=setup_requirements,
    tests_require=test_requirements,
    test_suite='tests',
    entry_points={
        'console_scripts': [
            'mosamatic-desktop=mosamaticdesktop.main:main',
        ],
    },
    version=os.environ['VERSION'],
    zip_safe=False,
)


from typing import Any


class Parameter:
    @classmethod
    def NAME(cls):
        return cls.__qualname__

    def __init__(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None) -> None:
        self._name = name
        self._labelText = labelText
        self._optional = optional
        self._visible = visible
        self._defaultValue = defaultValue
        self._value = None

    def name(self) -> str:
        return self._name

    def labelText(self) -> str:
        return self._labelText
    
    def optional(self) -> bool:
        return self._optional
    
    def visible(self) -> bool:
        return self._visible
    
    def defaultValue(self) -> Any:
        return self._defaultValue
    
    def value(self) -> Any:
        return self._value
    
    def setValue(self, value: Any) -> None:
        self._value = value

    def copy(self):
        raise NotImplementedError()
    
    def toDict(self):
        data = {
            'name': self.name(),
            'labelText': self.labelText(),
            'optional': self.optional(),
            'visible': self.visible(),
            'defaultValue': self.defaultValue(),
            'value': self.value(),            
        }
        return data


from mosamaticdesktop.tasks.taskwidget import TaskWidget
from mosamaticdesktop.tasks.decompressdicomtask.decompressdicomtask import DecompressDicomTask


class DecompressDicomTaskWidget(TaskWidget):
    def __init__(self) -> None:
        super(DecompressDicomTaskWidget, self).__init__(taskType=DecompressDicomTask)

    def validate(self) -> None:
        pass


from PySide6.QtWidgets import QWidget, QComboBox

from mosamaticdesktop.data.datamanager import DataManager


class FileSetComboBox(QComboBox):
    def __init__(self, parent: QWidget=None) -> None:
        super(FileSetComboBox, self).__init__(parent)
        self._dataManager = DataManager()
        self.loadItems()

    def loadItems(self) -> None:
        self.addItem(None)
        fileSets = self._dataManager.fileSets()
        for fileSet in fileSets:
            self.addItem(fileSet.name())

    def showPopup(self) -> None:
        self.clear()
        self.loadItems()
        return super().showPopup()



import numpy as np
import statistics

from totalsegmentator.map_to_binary import class_map
from scipy.ndimage import binary_dilation
from skimage.morphology import binary_erosion, ball
from scipy.ndimage import label as label_ndimage
from collections import Counter

from mosamaticdesktop.logger import Logger

LOGGER = Logger()


class CheckSegmentation:
    def __init__(self, segmentation, scanName: str):
        self._segmentation = segmentation.get_fdata().astype(int)
        self._scanName = scanName

    def checkVertebraeMasksConsistency(self, seg, radius=2.5):
        try:
            classes_dic = class_map['total']

            # Apply morphological operations to the mask
            # This is done to seperate the segmentation masks of each vertebrae
            struct_elem = ball(radius)  # choose appropriate size and shape of structuring element

            new_seg = np.zeros_like(seg)

            spine = {}

            LOGGER.info("Applying morphological transformations...")

            for label_value in range(18, 42):
                # Create a binary array for the current label value
                # Morphological operations are applied one label at a time
                binary_array = (seg == label_value)
                if np.sum(binary_array) > 0:
                    spine[label_value] = np.sum(binary_array)
                else:
                    continue
                # Apply erosion to the segmentation array
                eroded_array = binary_erosion(binary_array, struct_elem)
                # Apply dilation to the eroded segmentation array
                dilated_array = binary_dilation(eroded_array, struct_elem)
                # Create a binary mask of the dilated segmentation array
                new_seg[dilated_array] = label_value

            LOGGER.info("Separating each 3D mask...")

            # We only keep the labels of the vertebrae
            mask = np.logical_or(seg < 18, seg > 41)
            new_seg[mask] = seg[mask]
            s = new_seg
            # print("mask", mask.shape)
            # print("seg", seg.shape)
            # print("seg[mask]", seg[mask].shape)
            # print("new_seg[mask]", new_seg[mask].shape)
            # with open('seg.npy', 'wb') as f:
            #     np.save(f, seg)
            # with open('new_seg.npy', 'wb') as f:
            #     np.save(f, new_seg)

            # We compute the average size of a vertebra mask
            # This is to distinguish between masks of full vertebrae or "isolated islands"
            # We trim some values to not include the extremes in the average
            # calculate the number of values to trim from both ends

            trim_percentage = 0.1 # define the percentage of values to trim
            trim_count = round(len(spine) * trim_percentage)
            # print("len(spine)", len(spine))
            if trim_count < 1: # trim at least the min and max val
                trim_count = 1
            # print("trim_count", trim_count)
            # sort the values
            sorted_values = sorted(spine.values())
            # print("sorted_values", sorted_values)
            # trim the values
            trimmed_values = sorted_values[trim_count:len(sorted_values) - trim_count]
            # calculate the mean of the trimmed values
            if trimmed_values:
                trimmed_mean = statistics.mean(trimmed_values)
            else:
                trimmed_mean = 0
            # print("trimmed_mean", trimmed_mean)

            mask_arr = np.logical_and(s >= 18, s <= 41)

            # Perform connected component analysis on the mask
            labels, num_labels = label_ndimage(mask_arr)
            # print("num_labels", num_labels)
            # print("labels", labels.shape)

            label_assigned = {}

            LOGGER.info('Checking for inconsistencies in labeling ...')

            # Check each identified 3D mask
            for i in range(1, num_labels + 1):
                mask = (labels == i)
                indices = np.argwhere(mask)
                center = np.mean(indices, axis=0)
                vertebra_labels = s[mask]
                # Count each unique label in the 3D label
                label_counts = Counter(vertebra_labels)
                if len(label_counts) > 1:
                    # Mask contains pixels from multiple vertebrae labels, keep only the largest connected component for the correct label
                    correct_label = max(label_counts, key=label_counts.get)
                    LOGGER.info("Multiple labels are assigned to the same segmentation mask as {}".format(str(classes_dic[correct_label])))
                else:
                    correct_label = list(label_counts.keys())[0]
                # if the label has been previously assigned, we detect what the most plausible error is
                if correct_label in label_assigned:
                    if label_counts[correct_label] < trimmed_mean / 4:
                        LOGGER.info("{} appear more than once. A small seperate section is given this label.".format(str(classes_dic[correct_label])))
                    else:
                        if center[0] < label_assigned[correct_label]:
                            if correct_label < 18:
                                LOGGER.info("{} appear more than once. It might have needed to be assigned the saccrum label.".format(str(classes_dic[correct_label])))
                            else:
                                LOGGER.info("{} appear more than once. It might have needed to be assigned the previous vertebra label.".format(str(classes_dic[correct_label])))
                        else:
                            LOGGER.info("{} appear more than once. It might have needed to be assigned the following vertebra label.".format(str(classes_dic[correct_label])))
                else:
                    # Add the current label to the list of assigned labels
                    label_assigned[correct_label] = center[0]
                LOGGER.info("Check Done")
            return True
        except Exception as e:
            LOGGER.error("Error: {}".format(e))
        return False

    def checkOrientation(self, seg, precision=5):
        wrong_orientation = False

        # Each organ and bones is assigned a level to be able to check wheter the segmentation appear in the correct order
        map_organs_to_level = {'brain': 1, 'heart_myocardium': 2, 'heart_atrium_left': 2, 'heart_atrium_right': 2,
                            'heart_ventricle_left': 2
            , 'heart_ventricle_right': 2, 'stomach': 3, 'urinary_bladder': 4}
        map_bones_to_lvl = {'vertebrae_C2': 2, 'vertebrae_C4': 4, 'vertebrae_C6': 6, 'vertebrae_T1': 8, 'vertebrae_T3': 10
            , 'vertebrae_T5': 12, 'vertebrae_T7': 14, 'vertebrae_T9': 16, 'vertebrae_T11': 18, 'vertebrae_L1': 20,
                            'vertebrae_L3': 22
            , 'vertebrae_L5': 24, 'hip_right': 24, 'hip_left': 24, 'sacrum': 24}

        LOGGER.info("Iterating through slices ...")

        classes_dic = class_map["total"]
        # First we loop through the scan to check the organs are correct then we check the bones
        for i in range(2):
            if i == 0:
                map_classes_to_lvl = map_organs_to_level
            if i == 1:
                map_classes_to_lvl = map_bones_to_lvl
            start_index = len(seg[0][0])
            # print("start_index", start_index)
            # print("seg", seg.shape)
            map_lvl_to_classes = {}
            for k, v in map_classes_to_lvl.items():
                map_lvl_to_classes[v] = map_lvl_to_classes.get(v, []) + [k]
            # We loop every 5 slice, for each slice we keep in memory the segmentation that is supposed to be at the "highest level"
            length = seg.shape[2]
            for y in range(length):
                if start_index - precision > precision:
                    start_index = start_index - precision
                    # print("start_index", start_index)
                    slice = seg[:, :, start_index]
                    # slice = seg[start_index]
                    # from the slice we store every unique label
                    count = Counter(slice.flatten())
                    slice = set(slice.flatten())
                    # we remove no label ie 0
                    slice.remove(0)
                    arr = []
                    for el in slice:
                        if count[el] > 20:
                            cat = classes_dic[el]
                            arr.append(cat)
                    # we only keep labels that have been predefined a level to check if they appear in the correct order
                    arr = [i for i in arr if i in map_classes_to_lvl]
                    if len(arr) > 0:
                        # we find out of each label which is at max assigned level
                        max_lvl1 = 1
                        for cat in arr:
                            if map_classes_to_lvl[cat] > max_lvl1:
                                max_lvl1 = map_classes_to_lvl[cat]
                        index = start_index
                        # We loop every x (precision) slice appearing after the slice being currently checked. To make sure every segmentation is at a lower or equal level as the max level identified in slice 1
                        for x in range(start_index - precision):
                            # we repeat the same process on the following slices and compare to max_lvl1
                            if index - precision > 0:
                                index = index - precision
                                slice = seg[:, :, index]
                                # slice = seg[index]
                                count = Counter(slice.flatten())
                                slice = set(slice.flatten())
                                slice.remove(0)
                                arr = []
                                for el in slice:
                                    if count[el] > 20:
                                        cat = classes_dic[el]
                                        arr.append(cat)
                                arr = [i for i in arr if i in map_classes_to_lvl]
                                if len(arr) > 0:
                                    max_lvl2 = 1
                                    for cat in arr:
                                        if map_classes_to_lvl[cat] > max_lvl2:
                                            max_lvl2 = map_classes_to_lvl[cat]
                                    # If 1 of the segmentation is at a lower level, we print a warning of this
                                    if max_lvl2 < max_lvl1:
                                        wrong_orientation = True
                                        LOGGER.info("{}".format(map_lvl_to_classes[max_lvl2]))
                                        LOGGER.info('appear after')
                                        LOGGER.info("{}".format(map_lvl_to_classes[max_lvl1]))
                                        LOGGER.info('----')
                                        break
                        if wrong_orientation:
                            break
        if wrong_orientation:
            LOGGER.info("Input CT scan may not be in the correct orientation")
            return False
        return True

    def execute(self):
        if self.checkOrientation(self._segmentation):
            if self.checkVertebraeMasksConsistency(self._segmentation):
                return True
        return False


from mosamaticdesktop.tasks.taskwidget import TaskWidget
from mosamaticdesktop.tasks.copyfilesettask.copyfilesettask import CopyFileSetTask


class CopyFileSetTaskWidget(TaskWidget):
    def __init__(self) -> None:
        super(CopyFileSetTaskWidget, self).__init__(taskType=CopyFileSetTask)
    
    def validate(self) -> None:
        pass


import os
import pydicom
import pydicom.errors
import numpy as np


class SliceSelector:
    def __init__(self, roi, volume, dicomDirectory):
        self._roi = roi
        self._volume = volume
        self._dicomDirectory = dicomDirectory

    @staticmethod
    def get_min_max_slice_idx(roi):
        roi_data = roi.get_fdata()
        nr_slices = roi_data.shape[2]
        i_min = -1
        for i in range(nr_slices):
            slice = roi_data[:,:,i]
            if 1 in np.unique(slice):
                i_min = i
                break
        i_max = -1
        for i in range(nr_slices):
            slice = roi_data[:,:,nr_slices-i-1]
            if 1 in np.unique(slice):
                i_max = nr_slices - i
                break
        return i_min, i_max

    @staticmethod
    def get_z_coord_patient_position(i, volume):
        M = volume.affine[:3, :3]
        abc = volume.affine[:3, 3]
        return (M.dot([0, 0, i]) + abc)[2]

    def get_min_max_z_coord_patient_position(self, i_min, i_max, volume):
        z_min = self.get_z_coord_patient_position(i_min, volume)
        z_max = self.get_z_coord_patient_position(i_max, volume)
        return z_min, z_max

    @staticmethod
    def get_dicom_z(file_path):
        try:
            p = pydicom.dcmread(file_path, stop_before_pixels=True)            
            return p.ImagePositionPatient[2]
        except pydicom.errors.InvalidDicomError:
            pass
        return None

    def get_dicom_images_between(self, z_min, z_max, dicom_directory):
        z_coords = {}
        for f in os.listdir(dicom_directory):            
            f_path = os.path.join(dicom_directory, f)
            z = self.get_dicom_z(f_path)
            if z:
                z_coords[z] = f_path
        z_min_nearest = min(list(z_coords.keys()), key=lambda x:abs(x - z_min))
        z_max_nearest = min(list(z_coords.keys()), key=lambda x:abs(x - z_max))
        file_paths = []
        for z in z_coords.keys():
            if z_min_nearest <= z <= z_max_nearest:
                file_paths.append(z_coords[z])
        return file_paths

    def execute(self):
        i_min, i_max = self.get_min_max_slice_idx(self._roi)
        z_min, z_max = self.get_min_max_z_coord_patient_position(i_min, i_max, self._volume)
        z_median = z_min + np.abs(z_max - z_min) * 0.50
        output_files = self.get_dicom_images_between(z_median, z_median, self._dicomDirectory)
        return output_files


from typing import Any

from PySide6.QtWidgets import QWidget, QVBoxLayout, QLabel, QLineEdit, QVBoxLayout

from mosamaticdesktop.tasks.parameter import Parameter
from mosamaticdesktop.tasks.parameterwidget import ParameterWidget


class TextParameterWidget(ParameterWidget):
    def __init__(self, parameter: Parameter, parent: QWidget=None) -> None:
        super(TextParameterWidget, self).__init__(parameter=parameter, parent=parent)
        if self.parameter().defaultValue() is not None:
            self.parameter().setValue(self.parameter().defaultValue())
        self._pathLineEdit = None
        self.initUi()

    def initUi(self) -> None:
        self._pathLineEdit = QLineEdit(self)
        self._pathLineEdit.setText(self.parameter().defaultValue())
        self._pathLineEdit.textChanged.connect(self.pathChanged)
        self.layout().addWidget(QLabel(self.parameter().labelText()))
        self.layout().addWidget(self._pathLineEdit)

    def pathChanged(self, text) -> None:
        self.parameter().setValue(text)